#!/usr/bin/env python

"""
Define regions of interest with napari and save as roi files.

With this script you can choose a file name, open the localization file in napari.
Draw regions of interest as additional shapes in napari.
Upon closing napari each shape is taken as single roi and saved as _roi.yaml file.

To run the script::

    locan rois -f <file> -t <file type> -i <roi file indicator> --bin_size <>
    --rescale <string with tuple or rescale>

Try for instance::

    locan rois -f "locan/tests/test_data/five_blobs.txt" -t 1 -i "_roi" --bin_size 10

See Also
--------
locan.data.rois.select_by_drawing_napari
"""
import re
import argparse
from pathlib import Path

from locan.dependencies import HAS_DEPENDENCY
from locan.constants import FileType
from locan.gui.io import file_dialog
import locan.locan_io.locdata.io_locdata as io
from locan.render.render2d import select_by_drawing_napari

if HAS_DEPENDENCY["napari"]: import napari


def sc_draw_roi_napari(file_path=None, file_type=FileType.CUSTOM, roi_file_indicator='_roi', **kwargs):
    """
    Define regions of interest by drawing a boundary.

    Parameters
    ----------
    file_path : str, bytes, os.PathLike
        File path to localization data.
    file_type : int, str, locan.constants.FileType, metadata_pb2
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    roi_file_indicator : str
        Indicator to add to the localization file name and use as roi file name (with further extension .yaml).
    kwargs : dict
        Other parameters passed to :func:`render_2d_napari`.
    """
    # choose file interactively
    if file_path is None:
        file_path = Path(file_dialog(message='choose file', filter='*.txt; *.csv')[0])
    else:
        file_path = Path(file_path)

    print(file_path)

    # load data
    dat = io.load_locdata(path=file_path, file_type=file_type)

    # set roi
    rois = select_by_drawing_napari(locdata=dat, **kwargs)
    napari.run()
    print(rois)

    # save roi
    roi_path_list = []
    for i, roi in enumerate(rois):
        roi_file = file_path.stem + roi_file_indicator + f'_{i}.yaml'
        roi_path = file_path.with_name(roi_file)
        roi_path_list.append(roi_path)
        roi.to_yaml(path=roi_path)

    return roi_path_list


def type_converter_rescale(input_string):
    if input_string == 'None':
        return None
    elif input_string == 'True':
        return True
    elif input_string == 'False':
        return False
    else:
        pattern = re.match(r'\(?([0-9]*[.]?[0-9]+),?\s?([0-9]*[.]?[0-9]+)\)?', input_string)
        if pattern:
            return tuple(float(element) for element in pattern.groups())
        else:
            return input_string


def _add_arguments(parser):
    parser.add_argument('-f', '--file', dest='file', type=str, default=None,
                        help='File path to localization data.')
    parser.add_argument('-t', '--type', dest='type', type=int, default=2,
                        help='Integer or string indicating the file type.')
    parser.add_argument('-i', '--indicator', dest='roi_file_indicator', type=str, default='_roi',
                        help='Indicator to add to the localization file name and use as roi file name '
                             '(with further extension .yaml).')
    parser.add_argument('--bin_size', dest='bin_size', type=float, default=50,
                        help='Keyword passed to render function.')
    parser.add_argument('--rescale', dest='rescale', type=type_converter_rescale, default='EQUALIZE',
                        help='Rescale intensity values. Keyword passed to render_2d_napari function.')


def main(args=None):

    parser = argparse.ArgumentParser(description='Set roi by drawing a boundary.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_draw_roi_napari(returned_args.file, returned_args.type, returned_args.roi_file_indicator,
                       bin_size=returned_args.bin_size, rescale=returned_args.rescale)


if __name__ == '__main__':
    main()



"""
Analyze cross dependencies between localization properties.

Analyze cross dependencies as indicated by the correlation coefficients between any two localization properties.
"""
from collections import namedtuple
import logging

import numpy as np
import matplotlib.pyplot as plt
from lmfit import Model, Parameters

from locan.render.render2d import render_2d_mpl
from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.render.render2d import histogram
from locan.configuration import COLORMAP_DIVERGING


__all__ = ['LocalizationPropertyCorrelations']

logger = logging.getLogger(__name__)


##### The algorithms

def _localization_property_correlations(locdata, loc_properties=None):
    if loc_properties is None:
        results = locdata.data.corr()
    else:
        results = locdata.data[loc_properties].corr()
    return results

##### The specific analysis classes

class LocalizationPropertyCorrelations(_Analysis):
    """
    Compute and analyze correlation coefficients between any two localization properties.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_properties : list, None
        Localization properties to be analyzed. If None all are used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        The correlation coefficients..
    """
    def __init__(self, meta=None, loc_properties=None):
        super().__init__(meta=meta, loc_properties=loc_properties)
        self.results = None

    def compute(self, locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _localization_property_correlations(locdata=locdata, **self.parameter)
        return self

    def report(self):
        if not self:
            logger.warning('No results available')
            return

        print('Fit results for:\n')
        print(self.results.model_result.fit_report(min_correl=0.25))
        # print(self.results.fit_results.best_values)


    def plot(self, ax=None, cbar=True, colorbar_kws=None, **kwargs):
        """
        Provide heatmap of all correlation values as :class:`matplotlib.axes.Axes` object.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        cbar : bool
            If true draw a colorbar.
        colorbar_kws : dict
            Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.imshow`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        im = ax.imshow(self.results, **dict(dict(vmin=-1, vmax=1, cmap=COLORMAP_DIVERGING), **kwargs))
        columns = self.results.columns

        # We want to show all ticks...
        ax.set_xticks(np.arange(len(columns)))
        ax.set_yticks(np.arange(len(columns)))

        # ensure correct scaling
        ax.set_xticks(np.arange(len(columns) + 1) - .5, minor=True)
        ax.set_yticks(np.arange(len(columns) + 1) - .5, minor=True)
        ax.tick_params(which="minor", bottom=False, left=False)

        # ... and label them with the respective list entries
        ax.set_xticklabels(columns)
        ax.set_yticklabels(columns)

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(len(columns)):
            for j in range(len(columns)):
                text = ax.text(j, i, round(self.results.values[i, j], 2),
                               ha="center", va="center", color="w")

        ax.set_title("Localization Property Correlations")

        # Create colorbar
        if cbar:
            if colorbar_kws is None:
                cbar = ax.figure.colorbar(im)
            else:
                cbar = ax.figure.colorbar(im, **colorbar_kws)
            # cbar.ax.set_ylabel('correlation', rotation=-90, va="bottom")

        ax.figure.tight_layout()
        return ax



# -*- coding: utf-8 -*-

# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
import asyncio

# change event loop in windows for python 3.8
if sys.version_info.major == 3 and sys.version_info.minor == 8 and sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

# General information about the project.
project = 'Locan'
author = 'Locan Developers'
copyright = '2018-2022, Locan Developers'

# Version number generated by setuptools_scm
from importlib.metadata import version
release = version('locan')
# take major/minor
# version = '.'.join(release.split('.')[:2])
version = release


# -- General configuration ---------------------------------------------------

# Minimal Sphinx version needed for documentation
# needs_sphinx = '1.0'

# Sphinx extension modules
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'IPython.sphinxext.ipython_console_highlighting',
    'IPython.sphinxext.ipython_directive',
    'myst_nb',
    # 'sphinx_rtd_theme',
    'sphinx_copybutton'
    ]

# autosummary settings
# Make _autosummary files and include them
autosummary_generate = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True

# myST-NB settings
nb_execution_mode = 'force'  # 'off' 'cache'

# copybutton settings
copybutton_prompt_text = ">>> "

# The language for content autogenerated by Sphinx.
language = None

# The suffixes of source filenames.
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'myst-nb',
    '.ipynb': 'myst-nb',
    '.myst': 'myst-nb',
}

# The master/root toctree document.
master_doc = root_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'jupyter_execute', '.jupyter_cache', '**.ipynb_checkpoints']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
# html_theme = "sphinx_rtd_theme"
html_theme = "furo"

# Theme options
# for furo
html_theme_options = {
    # 'sidebar_hide_name': True,
}

html_title = f'{project}\n{release}'
# html_title = f'{release}'
html_short_title = f'{project}'

html_logo = '_static/logo.png'
html_favicon = '_static/favicon.ico'

html_static_path = ['_static']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f'{project}_docs'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '11pt',

    # Additional stuff for the LaTeX preamble.
    'preamble': '\setcounter{tocdepth}{2}',

    # Latex figure (float) alignment
    # 'figure_align': 'htbp',

    'classoptions': ',openany,oneside',
}

latex_logo = '_static/logo.png'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Locan.tex', 'Locan Documentation',
     author, 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'locan', 'Locan Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Locan', 'Locan Documentation',
     author, 'Locan', 'One line description of project.',
     'Miscellaneous'),
]



from pathlib import Path
import tempfile
import pickle

import pytest

from locan.analysis.pipeline import Pipeline, computation_test
from locan.analysis import metadata_analysis_pb2


def test_Pipeline(locdata_2d):
    with pytest.raises(TypeError):
        Pipeline()
    with pytest.raises(TypeError):
        Pipeline(computation=None, parameter='my_test')

    pipe = Pipeline(computation=computation_test, parameter='my_test')
    assert pipe
    assert isinstance(pipe.meta, metadata_analysis_pb2.AMetadata)
    assert pipe.kwargs == dict(parameter='my_test')
    pipe.compute()
    assert pipe.test == 'my_test'

    # this is not recommended since self.parameters are not updated automatically.
    computation_test(pipe, parameter='my_next_test')
    assert pipe.test == 'my_next_test'
    assert pipe.kwargs == dict(parameter='my_test')

    # several parameter including locdata reference and piped compute method.
    pipe = Pipeline(computation=computation_test, locdata=locdata_2d, parameter='my_test').compute()
    assert pipe.locdata is locdata_2d
    assert pipe.parameter['locdata'] is locdata_2d
    assert pipe.test == 'my_test'

    # print(pipe.computation_as_string())
    assert type(pipe.computation_as_string()) is str

    # save compute as text
    with tempfile.TemporaryDirectory() as tmp_directory:
        # for visual inspection use:
        # file_path = ROOT_DIR / 'tests/test_data/pipe.txt'
        file_path = Path(tmp_directory) / 'pipe.txt'
        pipe.save_computation(file_path)
        with open(file_path) as f:
            first_line = f.readline()
            assert first_line == "Analysis Pipeline: Pipeline\n"

    # pickling
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_pipeline.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(pipe, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            pipe_pickled = pickle.load(file)
        assert len(pipe_pickled.parameter) == len(pipe.parameter)
        assert isinstance(pipe_pickled.meta, metadata_analysis_pb2.AMetadata)
        assert pipe_pickled.meta == pipe.meta
        assert pipe_pickled.test == 'my_test'



import pytest
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt  # this import is required for visual inspection

from locan import LocData
from locan.analysis.localization_precision import PairwiseDistance1d, PairwiseDistance1dIdenticalSigmaZeroMu, \
    PairwiseDistance2d, PairwiseDistance2dIdenticalSigma, PairwiseDistance2dIdenticalSigmaZeroMu, \
    PairwiseDistance3d, PairwiseDistance3dIdenticalSigmaZeroMu, \
    LocalizationPrecision, _DistributionFits


@pytest.fixture()
def locdata_simple_1d():
    locdata_dict = {
        'position_y': range(10),
        'frame': [1, 1, 1, 2, 2, 3, 5, 8, 9, 10]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_simple_2d():
    locdata_dict = {
        'position_x': range(10),
        'position_y': range(10),
        'frame': [1, 1, 1, 2, 2, 3, 5, 8, 9, 10]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_simple_3d():
    locdata_dict = {
        'position_x': range(10),
        'position_y': range(10),
        'position_z': range(10),
        'frame': [1, 1, 1, 2, 2, 3, 5, 8, 9, 10]
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_PairwiseDistance1d():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance1d()
    assert distribution.shapes == 'mu, sigma_1, sigma_2'
    pdf_values = distribution.pdf(x=x_data, mu=1, sigma_1=0.1, sigma_2=0.2)
    assert pdf_values == pytest.approx([1.61998219e-04, 1.78412412e+00, 8.09991096e-05])


def test_PairwiseDistance1dIdenticalSigmaZeroMu():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance1dIdenticalSigmaZeroMu()
    assert distribution.shapes == 'sigma'
    pdf_values = distribution.pdf(x=x_data, sigma=0.1)
    assert pdf_values == pytest.approx([7.97884561e+00, 1.53891973e-21, 1.10418967e-86])


def test_PairwiseDistance2d():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance2d()
    assert distribution.shapes == 'mu, sigma_1, sigma_2'
    pdf_values = distribution.pdf(x=x_data, mu=1, sigma_1=0.1, sigma_2=0.2)
    assert pdf_values == pytest.approx([0.00000000e+00, 1.79560624e+00, 1.14913178e-04])


def test_PairwiseDistance2dIdenticalSigma():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance2dIdenticalSigma()
    assert distribution.shapes == 'mu, sigma'
    pdf_values = distribution.pdf(x=x_data, mu=1, sigma=0.1)
    assert pdf_values == pytest.approx([0.00000000e+00, 3.99443793e+00, 1.08886261e-21])


def test_PairwiseDistance2dIdenticalSigmaZeroMu():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance2dIdenticalSigmaZeroMu()
    assert distribution.shapes == 'sigma'
    pdf_values = distribution.pdf(x=x_data, sigma=0.1)
    assert pdf_values == pytest.approx([0.00000000e+00, 1.92874985e-20, 2.76779305e-85])


def test_PairwiseDistance3d():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance3d()
    assert distribution.shapes == 'mu, sigma_1, sigma_2'
    pdf_values = distribution.pdf(x=x_data, mu=1, sigma_1=0.1, sigma_2=0.2)
    assert pdf_values == pytest.approx([0.00000000e+00, 1.78412412e+00, 1.61998219e-04])


def test_PairwiseDistance3dIdenticalSigmaZeroMu():
    x_data = np.array([0, 1, 2])
    distribution = PairwiseDistance3dIdenticalSigmaZeroMu()
    assert distribution.shapes == 'sigma'
    pdf_values = distribution.pdf(x=x_data, sigma=0.1)
    assert pdf_values == pytest.approx([0.00000000e+00, 1.53891973e-19, 4.41675869e-84])


def test_LocalizationPrecision_empty(caplog):
    lp = LocalizationPrecision().compute(LocData())
    lp.fit_distributions()
    lp.plot()
    lp.hist()
    assert caplog.record_tuples == [('locan.analysis.localization_precision', 30, 'Locdata is empty.'),
                                    ('locan.analysis.localization_precision', 30, 'No results available to be fitted.')]


def test_Localization_precision(locdata_simple_2d):
    # print(locdata_simple.data)
    lp = LocalizationPrecision().compute(locdata=locdata_simple_2d)
    # print(lp.results)
    for prop in ['position_delta_x', 'position_delta_y', 'position_distance']:
        assert(prop in lp.results.columns)


@pytest.mark.visual
def test_Localization_precision_plot(locdata_rapidSTORM_2d):
    lp = LocalizationPrecision().compute(locdata=locdata_rapidSTORM_2d)
    lp.plot(window=10)
    plt.show()
    print(lp.results)
    print(lp.meta)


def test_Distribution_fits_1d(locdata_simple_1d):
    # print(locdata_simple.data)
    lp = LocalizationPrecision().compute(locdata=locdata_simple_1d)
    distribution_statistics = _DistributionFits(lp)
    distribution_statistics.fit()
    assert('position_distance_sigma' in distribution_statistics.parameters)
    distribution_statistics.plot()
    # plt.show()

    lp.fit_distributions(loc_property=None)
    assert lp.distribution_statistics.position_delta_y_loc
    # print(lp.distribution_statistics)

    lp.fit_distributions(loc_property='position_delta_y', floc=0)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_delta_y_loc == 0

    lp.fit_distributions(loc_property='position_distance')
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 0

    lp.fit_distributions(loc_property='position_distance', floc=1)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 1

    plt.close('all')


def test_Distribution_fits(locdata_simple_2d):
    # print(locdata_simple.data)
    lp = LocalizationPrecision().compute(locdata=locdata_simple_2d)
    distribution_statistics = _DistributionFits(lp)
    distribution_statistics.fit()
    assert('position_distance_sigma' in distribution_statistics.parameters)
    distribution_statistics.plot()
    # plt.show()

    lp.fit_distributions(loc_property=None)
    assert lp.distribution_statistics.position_delta_x_loc
    # print(lp.distribution_statistics)

    lp.fit_distributions(loc_property='position_delta_x', floc=0)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_delta_x_loc == 0

    lp.fit_distributions(loc_property='position_distance')
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 0

    lp.fit_distributions(loc_property='position_distance', floc=1)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 1

    plt.close('all')


def test_Distribution_fits_3d(locdata_simple_3d):
    # print(locdata_simple.data)
    lp = LocalizationPrecision().compute(locdata=locdata_simple_3d)
    distribution_statistics = _DistributionFits(lp)
    distribution_statistics.fit()
    assert('position_distance_sigma' in distribution_statistics.parameters)
    distribution_statistics.plot()
    # plt.show()

    lp.fit_distributions(loc_property=None)
    assert lp.distribution_statistics.position_delta_x_loc
    # print(lp.distribution_statistics)

    lp.fit_distributions(loc_property='position_delta_x', floc=0)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_delta_x_loc == 0

    lp.fit_distributions(loc_property='position_distance')
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 0

    lp.fit_distributions(loc_property='position_distance', floc=1)
    assert lp.distribution_statistics.parameters
    assert lp.distribution_statistics.parameter_dict()
    assert lp.distribution_statistics.position_distance_loc == 1

    plt.close('all')


def test_Localization_precision_histogram(locdata_simple_2d):
    lp = LocalizationPrecision().compute(locdata=locdata_simple_2d)
    # lp.hist(fit=True)

    with pytest.raises(AttributeError):
        lp.hist(loc_property='position_delta_x', fit=False)
        assert lp.distribution_statistics.position_delta_x_loc

    lp.hist(loc_property='position_delta_x', fit=True)
    # print(lp.distribution_statistics.parameter_dict())
    assert lp.distribution_statistics.position_delta_x_loc
    assert lp.distribution_statistics.position_delta_x_scale

    plt.close('all')


# standard LocData fixtures

@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 1),
    ('locdata_2d', 6),
    ('locdata_non_standard_index', 6)
])
def test_standard_locdata_objects(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_non_standard_index,
        fixture_name, expected):
    dat = eval(fixture_name)
    LocalizationPrecision(radius=1).compute(locdata=dat)



"""

File input/output for localization data.

"""
from __future__ import annotations
import logging
from enum import Enum

import pandas as pd

from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, convert_property_names
from locan.locan_io.locdata.rapidstorm_io import load_rapidSTORM_file, load_rapidSTORM_track_file
from locan.locan_io.locdata.thunderstorm_io import load_thunderstorm_file
from locan.locan_io.locdata.elyra_io import load_Elyra_file
from locan.locan_io.locdata.nanoimager_io import load_Nanoimager_file
from locan.locan_io.locdata.asdf_io import load_asdf_file
from locan.locan_io.locdata.decode_io import load_decode_file
from locan.locan_io.locdata.smap_io import load_SMAP_file
from locan.locan_io.locdata.smlm_io import load_SMLM_file


__all__ = ['load_txt_file', 'load_locdata']

logger = logging.getLogger(__name__)


def load_txt_file(path, sep=',', columns=None, nrows=None, property_mapping=None, convert=True, **kwargs):
    """
    Load localization data from a txt file.

    Locan column names are either supplied or read from the first line header.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a localization file to load.
    sep : str
        separator between column values (Default: ',')
    columns : list of str, None
        Locan column names. If None the first line is interpreted as header (Default: None).
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded (Default: None).
    property_mapping : dict[str: str] or list[dict]
        Mappings between column names and locan property names
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    # define columns
    if columns is None:
        dataframe = pd.read_csv(path, sep=sep, nrows=nrows,
                                **dict(dict(skiprows=0), **kwargs))

        column_keys = convert_property_names(dataframe.columns, property_mapping=property_mapping)
        dataframe.columns = column_keys
    else:
        column_keys = convert_property_names(columns, property_mapping=property_mapping)
        dataframe = pd.read_csv(path, sep=sep, nrows=nrows,
                                **dict(dict(skiprows=1, names=column_keys), **kwargs))

    if convert:
        dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.CUSTOM
    dat.meta.file.path = str(path)

    del dat.meta.history[:]
    dat.meta.history.add(name='load_txt_file',
                         parameter='path={}, sep={}, columns={}, nrows={}'.format(path, sep, columns, nrows))

    return dat


def _map_file_type_to_load_function(file_type):
    """
    Interpret user input for file_type.

    Parameters
    ----------
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Identifier for the file type. Integer or string should be according to locan.constants.FileType.

    Returns
    -------
    callable
        Name of function for loading the localization file of `type`.
    """
    look_up_table = dict(
        load_txt_file=load_txt_file,
        load_rapidSTORM_file=load_rapidSTORM_file,
        load_Elyra_file=load_Elyra_file,
        load_thunderstorm_file=load_thunderstorm_file,
        load_asdf_file=load_asdf_file,
        load_Nanoimager_file=load_Nanoimager_file,
        load_rapidSTORM_track_file=load_rapidSTORM_track_file,
        load_SMLM_file=load_SMLM_file,
        load_decode_file=load_decode_file,
        load_SMAP_file=load_SMAP_file,
    )

    class LoadFunction(Enum):
        load_txt_file = 1
        load_rapidSTORM_file = 2
        load_Elyra_file = 3
        load_thunderstorm_file = 4
        load_asdf_file = 5
        load_Nanoimager_file = 6
        load_rapidSTORM_track_file = 7
        load_SMLM_file = 8
        load_decode_file = 9
        load_SMAP_file = 10

    try:
        if isinstance(file_type, int):
            function_name = LoadFunction(file_type).name
        elif isinstance(file_type, str):
            function_name = LoadFunction(locan.constants.FileType[file_type.upper()].value).name
        elif isinstance(file_type, locan.constants.FileType):
            function_name = LoadFunction(file_type.value).name
        elif isinstance(file_type, metadata_pb2):
            function_name = LoadFunction(file_type).name
        else:
            raise TypeError
        return look_up_table[function_name]
    except ValueError:
        raise ValueError(f'There is no load function for type {file_type}.')


def load_locdata(path, file_type=1, nrows=None, **kwargs):
    """
    Load data from localization file as specified by type.

    This function is a wrapper for read functions for the various types of SMLM data.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a localization data file to load.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    return _map_file_type_to_load_function(file_type)(path=path, nrows=nrows, **kwargs)



"""
Module to deal with required and optional dependencies.

Optional dependencies are defined in setup.cfg as [options.extras_require].

In any module that requires an optional dependency the import should be carried out on top:
    if HAS_DEPENDENCY["package"]: import package

Any function that makes use of the optional dependency should be decorated with
    @needs_package("package")


CONSTANTS

.. autosummary::
   :toctree: ./

   INSTALL_REQUIRES
   EXTRAS_REQUIRE
   IMPORT_NAMES
   HAS_DEPENDENCY
   QT_BINDINGS
"""
from __future__ import annotations

import os
from functools import wraps
import importlib.util
import importlib.metadata
from enum import Enum
import warnings
from collections.abc import Iterable
import re


__all__ = ["needs_package", "IMPORT_NAMES", "INSTALL_REQUIRES", "EXTRAS_REQUIRE", "HAS_DEPENDENCY",
           "QtBindings", "QT_BINDINGS"]


def _has_dependency_factory(
        packages: Iterable[str],
        import_names: dict[str, str] | None = None
) -> dict[str, bool]:
    if import_names is None:
        import_names = IMPORT_NAMES
    has_dependency = dict()
    for package in packages:
        key = import_names.get(package, package)
        value = importlib.util.find_spec(key) is not None
        has_dependency[key] = value
    return has_dependency


def needs_package(package, import_names=None, has_dependency=None):
    """
    Function that returns a decorator to check for optional dependency.

    Parameters
    ----------
    package : str
        Package or dependency name that needs to be imported.
    import_names : dict[str, str] | None
        Mapping of import names on package names.
    has_dependency : dict | None
        Dictionary with bool indicator if package (import name) is available.

    Returns
    -------
    callable
        A decorator that raises ImportError if package is not available.
    """
    if import_names is None:
        import_names = IMPORT_NAMES
    if has_dependency is None:
        has_dependency = HAS_DEPENDENCY
    import_name = import_names.get(package, package)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not has_dependency[import_name]:
                raise ImportError(f"Function {func} needs {import_name} which cannot be imported.")
            return func(*args, **kwargs)
        return wrapper

    return decorator


def _get_dependencies(package: str) -> tuple[set[str], set[str]]:
    """Read out all required and optional (extra) dependencies for locan (PyPI names)."""
    requires = importlib.metadata.requires(package)

    pattern = r"[\w-]+"
    required_dependencies = {re.match(pattern, item).group() for item in requires if "extra ==" not in item}
    extra_dependencies = {re.match(pattern, item).group() for item in requires if "extra ==" in item}

    return required_dependencies, extra_dependencies


#: List of required dependencies (PyPi package names)
INSTALL_REQUIRES = set()

#: List of optional dependencies (PyPi package names)
EXTRAS_REQUIRE = set()
INSTALL_REQUIRES, EXTRAS_REQUIRE = _get_dependencies(package='locan')

#: A dictionary mapping PyPi package names to import names if they are different
IMPORT_NAMES = dict()
IMPORT_NAMES["ruamel"] = "ruamel.yaml"
IMPORT_NAMES["fast-histogram"] = "fast_histogram"
IMPORT_NAMES["boost-histogram"] = "boost_histogram"
IMPORT_NAMES["protobuf"] = "google.protobuf"
IMPORT_NAMES["scikit-image"] = "skimage"
IMPORT_NAMES["scikit-learn"] = "sklearn"
IMPORT_NAMES["pytest-qt"] = "pytestqt"
IMPORT_NAMES["mpl-scatter-density"] = "mpl_scatter_density"

#: A dictionary indicating if dependency is available.
HAS_DEPENDENCY = _has_dependency_factory(packages=INSTALL_REQUIRES.union(EXTRAS_REQUIRE))


# Possible python bindings to interact with QT
class QtBindings(Enum):
    """
    Python bindings used to interact with Qt.
    """
    NONE = 'none'
    PYSIDE2 = 'pyside2'
    PYQT5 = 'pyqt5'


# Force python binding - only for testing purposes:
# os.environ['QT_API'] = QtBindings.PYQT5.value


# Determine Python bindings for QT interaction.
if 'QT_API' in os.environ:  # this is the case that QT_API has been set.
    try:
        QT_BINDINGS = QtBindings(os.environ['QT_API'])
    except KeyError:
        warnings.warn(f'The requested QT_API {os.environ["QT_API"]} cannot be imported. Will continue without QT.')
        QT_BINDINGS = QtBindings.NONE

    if QT_BINDINGS == QtBindings.PYSIDE2:
        if importlib.util.find_spec("PySide2") is None:
            warnings.warn(f'The requested QT_API {QT_BINDINGS.value} cannot be imported. Will continue without QT.')
            QT_BINDINGS = QtBindings.NONE

    elif QT_BINDINGS == QtBindings.PYQT5:
        if importlib.util.find_spec("PyQt5") is None:
            warnings.warn(f'The requested QT_API {QT_BINDINGS.value} cannot be imported. Will continue without QT.')
            QT_BINDINGS = QtBindings.NONE

else:  # this is the case that QT_API has not been set.
    if importlib.util.find_spec("PySide2") is not None:
        QT_BINDINGS = QtBindings.PYSIDE2
    elif importlib.util.find_spec("PyQt5") is not None:
        QT_BINDINGS = QtBindings.PYQT5
    else:
        QT_BINDINGS = QtBindings.NONE  # this is the case that no qt bindings are available.

# In order to force napari and other QT-using libraries to import with the correct Qt bindings
# the environment variable QT_API has to be set.
# See use of qtpy in napari which default to pyqt5 if both bindings are installed.
if QT_BINDINGS != QtBindings.NONE:
    os.environ['QT_API'] = QT_BINDINGS.value



"""

Deal with metadata in LocData objects.

Functions to modify metadata in LocData objects.

"""
from __future__ import annotations

import importlib
import tomli
import logging

from google.protobuf.message import Message
from google.protobuf import text_format, json_format

from locan.data import metadata_pb2


__all__ = ['metadata_to_formatted_string', 'metadata_from_toml', 'message_scheme']

logger = logging.getLogger(__name__)


def _modify_meta(locdata, new_locdata, function_name=None, parameter=None, meta=None):
    """
    Update metadata in Locdata after modification of locdata.

    Parameters
    ----------
    locdata : LocData
        original locdata before modification
    new_locdata : LocData
        new locdata after modification
    function_name : str
        Name of function that was applied for modification.
    parameter : dict
        Parameter for function that was applied for modification.
    meta : locan.data.metadata_pb2.Metadata
        Metadata about the current dataset and its history.

    Returns
    -------
    locan.data.metadata_pb2.Metadata
        Metadata about the current dataset and its history.
    """
    meta_ = metadata_pb2.Metadata()
    meta_.CopyFrom(locdata.meta)
    # try:
    #     meta_.ClearField("identifier")
    # except ValueError:
    #     pass
    #
    # try:
    #     meta_.ClearField("element_count")
    # except ValueError:
    #     pass
    #
    # try:
    #     meta_.ClearField("frame_count")
    # except ValueError:
    #     pass

    meta_.identifier = new_locdata.meta.identifier
    meta_.element_count = new_locdata.meta.element_count
    meta_.frame_count = new_locdata.meta.frame_count
    meta_.modification_time.CopyFrom(new_locdata.meta.creation_time)

    meta_.state = metadata_pb2.MODIFIED
    meta_.ancestor_identifiers.append(locdata.meta.identifier)
    meta_.history.add(name=function_name, parameter=str(parameter))

    if meta is None:
        pass
    elif isinstance(meta, dict):
        for key, value in meta.items():
            setattr(meta_, key, value)
    else:
        meta_.MergeFrom(meta)

    return meta_


def _dict_to_protobuf(dictionary: dict, message: Message, inplace: bool = False) -> Message | None:
    """
    Parse dictionary with message attributes and their values in message.
    """
    if inplace is False:
        message_ = message.__class__()
        message_.CopyFrom(message)
        message = message_

    for key, value in dictionary.items():
        try:
            attr_ = getattr(message, key)
        except AttributeError as e:
            logging.warning(f"AttributeError while parsing: {e}")
            break

        if isinstance(value, dict):
            _dict_to_protobuf(dictionary=value, message=attr_, inplace=True)
        elif isinstance(value, list):
            try:
                attr_.extend(value)
            except TypeError:
                for element in value:
                    submessage = attr_.add()
                    _dict_to_protobuf(dictionary=element, message=submessage, inplace=True)
        else:
            try:
                setattr(message, key, value)
            except AttributeError:
                if attr_.DESCRIPTOR.name == "Timestamp":
                    attr_.FromJsonString(value)
                elif attr_.DESCRIPTOR.name == "Duration":
                    attr_.FromNanoseconds(value)

    if inplace:
        return None
    else:
        return message


def metadata_to_formatted_string(message, **kwargs):
    """
    Get formatted string from Locdata.metadata.

    Parameters
    ----------
    message : google.protobuf.message.Message
        Protobuf message like locan.data.metadata_pb2.Metadata

    kwargs : dict
        Other kwargs that are passed to :func:`google.protobuf.text_format.MessageToString`.

    Returns
    -------
    str
        Formatted metadata string.
    """
    def message_formatter(message, indent: int, as_one_line: bool) -> str | None:
        if message.DESCRIPTOR.name in ["Timestamp", "Duration"]:
            return message.ToJsonString()
        else:
            return None

    return text_format.MessageToString(message, message_formatter=message_formatter, **kwargs)


def metadata_from_toml(file):
    """
    Turn toml file into protobuf message instances.

    Note
    -----
    Parses Timestamp elements from string '2022-05-14T06:58:00Z'.
    Parses Duration elements from int in nanoseconds.

    Parameters
    ----------
    file : str
        TOML file-like to load from.

    Returns
    -------
    dict[str, google.protobuf.message.Message]
        Message instances with name as declared in toml file.
    """
    # load toml file
    toml_dict = tomli.loads(file)

    # instantiate messages
    instances = {}
    for message in toml_dict.pop("messages"):
        module = importlib.import_module(message["module"])
        class_ = getattr(module, message["class_name"])
        instances[message["name"]] = class_()

    # parse values
    for message_name, dictionary in toml_dict.items():
        _dict_to_protobuf(dictionary=dictionary, message=instances[message_name], inplace=True)

    return instances


def message_scheme(message) -> dict:
    """
    Provide message scheme with defaults including nested messages.

    Parameters
    ----------
    message : google.protobuf.message.Message
        Protobuf message

    Returns
    -------
    dict
        A nested dictionary with all message fields including default values.
    """

    message_dict = json_format.MessageToDict(message, including_default_value_fields=True,
                                             preserving_proto_field_name=True)

    for descriptor in message.DESCRIPTOR.fields:
        if descriptor.type == descriptor.TYPE_MESSAGE:
            attr_ = getattr(message, descriptor.name)

            if descriptor.label != descriptor.LABEL_REPEATED:
                message_dict[descriptor.name] = message_scheme(attr_)

            elif descriptor.label == descriptor.LABEL_REPEATED \
                    and not 'ScalarMap' in type(attr_).__name__ \
                    and not 'MessageMapContainer' in type(attr_).__name__:
                attr_ = attr_.add()
                message_dict[descriptor.name] = message_scheme(attr_)

    return message_dict



"""

Aggregate localization data.

This module provides functions to bin LocData objects to form a histogram or image.

Specify bins through one of the parameters (`bins`, `bin_edges`, `n_bins`, `bin_size`, `bin_range`, `labels`)
as further outlined in the documentation for :class:`Bins`.
"""
from typing import Union
from collections.abc import Iterable
import warnings
from math import isclose
from collections import namedtuple

import numpy as np
import fast_histogram
import boost_histogram as bh

from locan.data.properties.locdata_statistics import ranges
from locan.data.locdata import LocData


__all__ = ['Bins', 'histogram']


def is_array_like(anything) -> bool:
    """
    Return true if `anything` can be turned into a numpy.ndarray without creating elements of type object.

    Catches numpy.VisibleDeprecationWarning.

    Parameters
    ----------
    anything
        Anything to be classified as being array-like or not.

    Returns
    -------
    bool
    """
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        try:
            return np.asarray(anything).dtype != object
        except np.VisibleDeprecationWarning as e:
            if "Creating an ndarray from ragged nested sequences" in str(e):
                return False
            else:
                raise e
        except Exception as e:
            raise e


def _is_scalar(element) -> bool:
    return isinstance(element, (int, float)) or \
           (is_array_like(element) and np.size(element) == 1 and np.ndim(element) == 0)


def _is_single_element(element) -> bool:
    return isinstance(element, (int, float)) or \
           (is_array_like(element) and np.size(element) == 1 and np.ndim(element) in (0, 1))


def _is_1d_array_of_scalar(element) -> bool:
    return is_array_like(element) and np.size(element) >= 1 and np.ndim(element) == 1


def _is_1d_array_of_two_or_more_scalar(element) -> bool:
    return is_array_like(element) and np.size(element) > 1 and np.ndim(element) == 1


def _is_1d_array_of_two_scalar(element) -> bool:
    return is_array_like(element) and np.size(element) == 2 and np.ndim(element) == 1


def _is_2d_homogeneous_array(element) -> bool:
    return is_array_like(element) and np.size(element) > 1 and np.ndim(element) == 2


def _is_2d_inhomogeneous_array_of_1d_array_of_scalar(element) -> bool:
    return not is_array_like(element) and all(_is_1d_array_of_scalar(el) for el in element)


def _is_2d_inhomogeneous_array(element) -> bool:
    return not is_array_like(element) and all(_is_scalar(el) or _is_1d_array_of_scalar(el) for el in element)


def _is_2d_array_of_1d_array_of_scalar(element) -> bool:
    return _is_2d_homogeneous_array(element) or _is_2d_inhomogeneous_array_of_1d_array_of_scalar(element)


"""
def _is_scalar(element) -> bool:
    return isinstance(element, (int, float)) or \
           (_is_array_like(element) and np.size(element) == 1 and np.ndim(element) == 0)


def _is_1d_array_of_scalar(element) -> bool:
    return _is_array_like(element) and np.size(element) > 0 and np.ndim(element) == 1


def _is_1d_array_of_two_scalar(element) -> bool:
    return _is_array_like(element) and np.size(element) == 2 and np.ndim(element) == 1


def _is_2d_homogeneous_array(element) -> bool:
    return _is_array_like(element) and np.size(element) > 1 and np.ndim(element) == 2


def _is_2d_inhomogeneous_array_of_1d_array_of_scalar(element) -> bool:
    return not _is_array_like(element) and len(element) > 1 and \
           all([((_is_array_like(element_) and np.ndim(element_) in (1, 2)) or
                  _is_scalar(element_)) for element_ in element])

"""


def _n_bins_to_bin_edges_one_dimension(n_bins, bin_range) -> np.ndarray:
    """
    Compute bin edges from n_bins and bin_range.

    Parameters
    ----------
    n_bins : int
        Number of bins.
    bin_range : tuple, list, numpy.ndarray of float with shape (2,)
        Minimum and maximum edge.

    Returns
    -------
    numpy.ndarray
        Array with bin edges.
    """
    return np.linspace(*bin_range, n_bins + 1, endpoint=True, dtype=float)


def _bin_size_to_bin_edges_one_dimension(bin_size, bin_range, extend_range=None) -> np.ndarray:
    """
    Compute bin edges from bin_size and bin_range.

    Use bin_edges if you need to use variable bin_sizes for Bins construction.

    Parameters
    ----------
    bin_size : float or tuple, list, numpy.ndarray of float with shape (n_bins,).
        One size or sequence of sizes for bins.
    bin_range : tuple, list, numpy.ndarray of float with shape (2,).
        Minimum and maximum edge of binned bin_range.
    extend_range : bool, None
        If for equally-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken from the input sequence (True).

    Returns
    -------
    numpy.ndarray
        Array of bin edges
    """
    if _is_scalar(bin_size):
        bin_edges = np.arange(*bin_range, bin_size, dtype=float)
        if bin_edges.size == 1:  # this is the case if bin_size is greater than the bin_range
            if extend_range is True:
                bin_edges = np.append(bin_edges, bin_edges + bin_size)
            elif extend_range is None or extend_range is False:
                bin_edges = np.array(bin_range)
            else:
                raise ValueError('`extend_range` must be None, True or False.')
        else:
            last_edge = bin_edges[-1] + bin_size
            if isclose(last_edge, bin_range[-1]):
                bin_edges = np.append(bin_edges, last_edge)
            else:
                if extend_range is None:
                    pass
                elif extend_range is True:
                    bin_edges = np.append(bin_edges, last_edge)
                elif extend_range is False:
                    bin_edges = np.append(bin_edges, bin_range[-1])
                else:
                    raise ValueError('`extend_range` must be None, True or False.')

    elif _is_1d_array_of_scalar(bin_size):
        if bin_size[0] > np.diff(bin_range):
            if extend_range is True:
                bin_edges = np.array([bin_range[0], bin_range[0] + bin_size[0]])
            elif extend_range is None or extend_range is False:
                bin_edges = bin_range
            else:
                raise ValueError('`extend_range` must be None, True or False.')
        else:
            bin_edges_ = np.concatenate((np.asarray([bin_range[0]]), np.cumsum(bin_size) + bin_range[0]))
            bin_edges = bin_edges_[bin_edges_ <= bin_range[-1]]
            if extend_range is None:
                pass
            elif extend_range is True:
                if len(bin_edges_) > len(bin_edges):
                    last_edge = bin_edges_[len(bin_edges)]
                    bin_edges = np.append(bin_edges, last_edge)
                else:
                    pass
            elif extend_range is False:
                bin_edges = np.append(bin_edges, bin_range[-1])
            else:
                raise ValueError('`extend_range` must be None, True or False.')

    else:
        raise TypeError("`bin_size` must be 0- or 1-dimensional.")

    return bin_edges


def _bin_edges_to_n_bins_one_dimension(bin_edges) -> int:
    n_bins = len(bin_edges) - 1
    return n_bins


def _bin_edges_to_n_bins(bin_edges):
    """
    Check if bins are equally sized and return the number of bins.

    Parameters
    ----------
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges)
        Array of bin edges for each dimension.

    Returns
    -------
    tuple of int
        Number of bins
    """
    if _is_1d_array_of_scalar(bin_edges):
        n_bins = (_bin_edges_to_n_bins_one_dimension(bin_edges),)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        n_bins = tuple(_bin_edges_to_n_bins_one_dimension(edges) for edges in bin_edges)
    else:
        raise TypeError('The shape of bin_edges must be (dimension, n_bin_edges).')
    return n_bins


def _bin_edges_to_bin_size_one_dimension(bin_edges) -> Union[float, np.ndarray]:
    differences = np.diff(bin_edges)
    if np.all(differences == differences[0]):
        bin_size = differences[0]
    else:
        bin_size = differences
    return bin_size


def _bin_edges_to_bin_size(bin_edges):
    """
    Compute the sizes of bins.

    Parameters
    ----------
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges)
        Array of bin edges for each dimension.

    Returns
    -------
    tuple of float or tuple of np.ndarray
        Bin sizes for each dimension
    """
    if _is_1d_array_of_scalar(bin_edges):
        bin_size = (_bin_edges_to_bin_size_one_dimension(bin_edges),)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        bin_size = tuple(_bin_edges_to_bin_size_one_dimension(edges) for edges in bin_edges)
    else:
        raise TypeError('The shape of bin_edges must be (dimension, n_bin_edges).')
    return bin_size


def _bin_edges_to_bin_centers(bin_edges):
    """
    Compute bin centers.

    Parameters
    ----------
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges)
        Array of bin edges for each dimension

    Returns
    -------
    tuple of numpy.ndarray of shape (dimension, n_bins)
        Bin centers
    """
    if _is_1d_array_of_scalar(bin_edges):
        bin_centers = (np.diff(bin_edges) / 2 + bin_edges[0:-1],)
    elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
        bin_centers = tuple(np.diff(edges) / 2 + edges[0:-1] for edges in bin_edges)
    else:
        raise TypeError('The shape of bin_edges must be (dimension, n_bin_edges).')
    return bin_centers


def _indices_to_bin_centers(bin_edges, indices):
    """
    Compute bin centers for given indices.

    Parameters
    ----------
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges)
        Array of bin edges for each dimension
    indices : int or tuple, list, numpy.ndarray of int with shape (n_indices, dimension)
        Array of multi-dimensional indices, e.g. reflecting a list of vertices.

    Returns
    -------
    numpy.ndarray of shape (n_indices, dimension)
        selected bin centers
    """
    bin_centers = _bin_edges_to_bin_centers(bin_edges)
    if len(bin_centers) == 1:
        bin_centers = bin_centers[0]
    indices = np.asarray(indices)

    if _is_1d_array_of_scalar(bin_centers):
        if _is_scalar(indices):
            selected_bin_centers = bin_centers[indices]
        elif _is_1d_array_of_scalar(indices):
            selected_bin_centers = bin_centers[indices]
        elif _is_2d_homogeneous_array(indices):
            selected_bin_centers = bin_centers[indices]
        else:
            raise TypeError('The given array shapes cannot be processed.')

    elif _is_2d_array_of_1d_array_of_scalar(bin_centers):
        if _is_scalar(indices):
            selected_bin_centers = [bc[indices] for bc in bin_centers]
        elif _is_1d_array_of_scalar(indices):
            selected_bin_centers = [bc[indices] for bc in bin_centers]
        elif _is_2d_homogeneous_array(indices):
            if len(bin_centers) != len(indices.T):
                raise TypeError("`bin_centers` and `indices` must be for the same dimension.")
            else:
                selected_bin_centers = np.array([bc[idx] for bc, idx in zip(bin_centers, indices.T)]).T
        else:
            raise TypeError('The given array shapes cannot be processed.')
    else:
        raise TypeError('The given array shapes cannot be processed.')

    return selected_bin_centers


class _BinsFromBoostHistogramAxis:
    """
    Adapter class for dealing with boost-histogram.axis elements through the `bins` parameter in :class:`Bins`.

    Parameters
    ----------
    bins : :class:`boost_histogram.axis.Axis` or :class:`boost_histogram.axis.AxesTuple`
    """

    def __init__(self, bins):
        if isinstance(bins, bh.axis.Axis):
            self._bins = bins
            self.dimension = 1
            self.n_bins = (self._bins.size,)
            self.bin_size = (tuple(self._bins.widths),)
            self.bin_edges = (self._bins.edges,)
            self.bin_range = ((self._bins.edges[0], self._bins.edges[-1]),)
            self.bin_centers = (self._bins.centers,)

        elif isinstance(bins, bh.axis.AxesTuple):
            self._bins = bins
            self.dimension = len(self._bins)
            self.n_bins = self._bins.size
            self.bin_size = tuple(tuple(arr) for arr in self._bins.widths.flatten())
            self.bin_edges = self._bins.edges.flatten()
            self.bin_range = tuple((axis.edges[0], axis.edges[-1]) for axis in self._bins)
            self.bin_centers = self._bins.centers.flatten()

        else:
            raise TypeError

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self._bins, attr)


class _BinsFromEdges:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    bin_edges : list, tuple or numpy.ndarray
        Array with bin edges for all or each dimension.
    """

    def __init__(self, bin_edges):

        if _is_1d_array_of_scalar(bin_edges):
            self.bin_edges = (np.array(bin_edges),)
            self.dimension = 1
            self.bin_range = ((bin_edges[0], bin_edges[-1]),)
        elif _is_2d_array_of_1d_array_of_scalar(bin_edges):
            self.bin_edges = tuple(np.array(edges) for edges in bin_edges)
            self.dimension = len(self.bin_edges)
            self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)
        else:
            raise TypeError("`bin_edges` must have 1 or 2 dimensions.")

        self.n_bins = _bin_edges_to_n_bins(self.bin_edges)
        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)


class _BinsFromNumber:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    n_bins : int or list, tuple or numpy.ndarray of ints
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_range : tuple, list, numpy.ndarray of float with shape (2,) or (dimension, 2)
        Minimum and maximum edge for all or each dimensions.
    """

    def __init__(self, n_bins, bin_range):

        if not is_array_like(n_bins) or np.ndim(n_bins) > 1:
            raise TypeError("`n_bins` must be 0- or 1-dimensional.")

        elif _is_scalar(n_bins):
            if _is_1d_array_of_two_scalar(bin_range):
                self.dimension = 1
                self.bin_edges = (_n_bins_to_bin_edges_one_dimension(n_bins, bin_range),)
                self.n_bins = (n_bins,)
                self.bin_range = (bin_range,)

            elif _is_2d_homogeneous_array(bin_range):
                self.dimension = len(bin_range)
                self.bin_edges = tuple(_n_bins_to_bin_edges_one_dimension(n_bins, single_range)
                                       for single_range in bin_range)
                self.n_bins = tuple(n_bins for _ in bin_range)
                self.bin_range = tuple(bin_range)

            else:
                raise TypeError('n_bins and/or bin_range have incorrect shapes.')

        elif _is_1d_array_of_scalar(n_bins):
            if _is_1d_array_of_two_or_more_scalar(bin_range):
                self.dimension = len(n_bins)
                self.bin_edges = tuple(_n_bins_to_bin_edges_one_dimension(n_bins=n, bin_range=bin_range)
                                       for n in n_bins)
                self.n_bins = tuple(n_bins)
                self.bin_range = tuple(bin_range for _ in n_bins)

            elif _is_2d_homogeneous_array(bin_range):
                if len(n_bins) != len(bin_range):
                    raise TypeError('n_bins and bin_range have incompatible shapes.')
                else:
                    self.dimension = len(n_bins)
                    self.bin_edges = tuple(_n_bins_to_bin_edges_one_dimension(n_bins=b, bin_range=r)
                                           for b, r in zip(n_bins, bin_range))
                    self.n_bins = tuple(n_bins)
                    self.bin_range = tuple(bin_range)

            else:
                raise TypeError('n_bins and/or bin_range have incorrect shapes.')

        else:
            raise TypeError('n_bins and/or bin_range have incorrect shapes.')

        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)


class _BinsFromSize:
    """
    Builder for :class:`Bins`.

    Parameters
    ----------
    bin_size : float, list, tuple or numpy.ndarray of float with shape (dimension,) or (dimension, n_bins)
        The size of bins for all or each bin and for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and (1, 3) for the other dimension.
    bin_range : tuple, list, numpy.ndarray of float with shape (2,) or (dimension, 2)
        Minimum and maximum edge for all or each dimensions.
    extend_range : bool or None
        If for equally-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken from the input sequence (True).
    """
    def __init__(self, bin_size, bin_range, extend_range=None):

        if _is_scalar(bin_size):
            if _is_1d_array_of_two_scalar(bin_range):
                self.dimension = 1
                self.bin_edges = (_bin_size_to_bin_edges_one_dimension(bin_size, bin_range, extend_range),)
                self.bin_range = (bin_range,)

            elif _is_2d_homogeneous_array(bin_range):
                self.dimension = len(bin_range)
                self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size, single_range, extend_range)
                                       for single_range in bin_range)
                self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

            else:
                raise TypeError('bin_size and/or bin_range have incorrect shapes.')

        elif _is_single_element(bin_size):
            if _is_1d_array_of_two_scalar(bin_range):
                self.dimension = 1
                self.bin_edges = (_bin_size_to_bin_edges_one_dimension(bin_size[0], bin_range, extend_range),)
                self.bin_range = (bin_range,)

            elif _is_2d_homogeneous_array(bin_range):
                self.dimension = len(bin_range)
                self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size[0], single_range, extend_range)
                                       for single_range in bin_range)
                self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

            else:
                raise TypeError('bin_size and/or bin_range have incorrect shapes.')

        elif _is_1d_array_of_scalar(bin_size):
            if _is_1d_array_of_two_scalar(bin_range):
                self.dimension = len(bin_size)
                self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size=bs, bin_range=bin_range,
                                                                            extend_range=extend_range)
                                       for bs in bin_size)
                self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

            elif _is_2d_homogeneous_array(bin_range):
                if len(bin_size) != len(bin_range):
                    raise TypeError('bin_size and bin_range have incompatible shapes.')
                else:
                    self.dimension = len(bin_size)
                    self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size=bs, bin_range=br,
                                                                                extend_range=extend_range)
                                           for bs, br in zip(bin_size, bin_range))
                    self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

            else:
                raise TypeError('bin_size and/or bin_range have incorrect shapes.')

        elif _is_2d_inhomogeneous_array(bin_size) or _is_2d_homogeneous_array(bin_size):  #_is_2d_array_of_1d_array_of_scalar(bin_size):
            if _is_1d_array_of_two_scalar(bin_range):
                self.dimension = len(bin_size)
                self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size=bs, bin_range=bin_range,
                                                                            extend_range=extend_range)
                                       for bs in bin_size)
                self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

            elif _is_2d_homogeneous_array(bin_range):
                if len(bin_size) != len(bin_range):
                    raise TypeError('bin_size and bin_range have incompatible shapes.')
                else:
                    self.dimension = len(bin_size)
                    self.bin_edges = tuple(_bin_size_to_bin_edges_one_dimension(bin_size=bs, bin_range=br,
                                                                                extend_range=extend_range)
                                           for bs, br in zip(bin_size, bin_range))
                    self.bin_range = tuple((edges[0], edges[-1]) for edges in self.bin_edges)

        elif np.ndim(bin_size) > 2:
            raise TypeError("`bin_size` must be 0-, 1- or 2-dimensional. \
                            Construct from bin_edges if you need variable bin_sizes in one dimension.")

        else:
            raise TypeError('bin_size and/or bin_range have incorrect shapes.')

        self.n_bins = _bin_edges_to_n_bins(self.bin_edges)
        self.bin_size = _bin_edges_to_bin_size(self.bin_edges)


# todo: add option for the following bin specifications.
# Bin specifications from an appropriate class or as defined in :func:`numpy.histogramdd`:
# The number of bins for all dimensions n.
# The number of bins for each dimension (nx, ny)
# A sequence of arrays ((edge_x1, edge_x2), (edge_y1, edge_y2)) describing the monotonically
# increasing bin edges along each dimension.
class Bins:
    """
    Bin definitions to be used in histogram and render functions.
    `Bins`can be instantiated from specifications for `bins` or `bin_edges`
    or for one of `n_bins` or `bin_size` in combination with `bin_range`.
    One and only one of (`bins`, `bin_edges`, `n_bins`, `bin_size`) must be different from None in any instantiating
    function.
    To pass bin specifications to other functions use an instance of `Bins` or `bin_edges`.

    Parameters
    ----------
    bins : `Bins` or `boost_histogram.axis.Axis` or None
        Specific class specifying the bins.

    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges) or None
        Array of bin edges for all or each dimension.

    bin_range : tuple, list, numpy.ndarray of float with shape (2,) or (dimension, 2)
        Minimum and maximum edge for all or each dimensions.

    n_bins : int, list, tuple or numpy.ndarray or None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.

    bin_size : float, list, tuple or numpy.ndarray of float with shape (dimension,) or (dimension, n_bins)
        The size of bins in units of locdata coordinate units for all or each bin and for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and (1, 3) for the other dimension.

    labels : list of str with shape (dimension,) or None
        Names for each bin axis.

    extend_range : bool or None
        Only applicable in combination with bin_size.
        If for equally-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range but all bins are equally-sized (None);
        the last bin_edge will be equal to the maximum bin_range but bins are not equally-sized (False);
        the last bin_edge will be larger than the maximum bin_range but all bins are equally-sized (True).

        If for variable-sized bins the final bin_edge is different from the maximum bin_range,
        the last bin_edge will be smaller than the maximum bin_range (None);
        the last bin_edge will be equal to the maximum bin_range (False);
        the last bin_edge will be larger than the maximum bin_range but taken from the input sequence (True).

    Attributes
    ----------
    dimension : int
        The number of dimensions for which bins are provided.

    bin_range : tuple of tuples of float with shape (dimension, 2)
        Minimum and maximum edge for each dimension.

    bin_edges : tuple of numpy.ndarray with shape (dimension,)
        Array(s) with bin edges for each dimension.

    n_bins : tuple of int with shape (dimension,)
        Number of bins for each dimension.

    bin_size : tuple of float with shape (dimension,) or tuple of numpy.ndarray with shape (dimension, n_bins)
        Size of bins for each dimension.

    bin_centers :  tuple of numpy.ndarray with shape (dimension,)
        Array(s) with bin centers for all or each dimension.

    labels : list of str or None
        Names for each bin axis.
    """

    def __init__(self, bins=None, n_bins=None, bin_size=None, bin_edges=None, bin_range=None,
                 labels=None, extend_range=None):

        # check for correct inputs
        excluding_parameter = (bins, n_bins, bin_size, bin_edges)
        excluding_parameter_strings = ('bins', 'n_bins', 'bin_size', 'bin_edges')
        n_inputs = sum(param is not None for param in excluding_parameter)
        if n_inputs != 1:
            raise ValueError(f"One and only one of {excluding_parameter_strings} must be different from None.")

        # inject builder class
        if bins is not None:
            if bin_range is not None:
                raise ValueError("The parameter `bin_range` is derived from bins class and must be None.")
            if isinstance(bins, Bins):
                self._bins = bins
            if isinstance(bins, (bh.axis.Axis, bh.axis.AxesTuple)):
                self._bins = _BinsFromBoostHistogramAxis(bins)
        elif n_bins is not None:
            self._bins = _BinsFromNumber(n_bins, bin_range)
        elif bin_size is not None:
            self._bins = _BinsFromSize(bin_size, bin_range, extend_range)
        elif bin_edges is not None:
            if bin_range is not None:
                raise ValueError("The parameter `bin_range` is derived from `bin_edges` and must be None.")
            self._bins = _BinsFromEdges(bin_edges)

        self._bin_centers = None
        self.labels = labels

    @property
    def dimension(self):
        return self._bins.dimension

    @property
    def bin_edges(self):
        return self._bins.bin_edges

    @property
    def n_bins(self):
        return self._bins.n_bins

    @property
    def bin_size(self):
        return self._bins.bin_size

    @property
    def bin_range(self):
        return self._bins.bin_range

    @property
    def bin_centers(self):
        if self._bin_centers is None:
            self._bin_centers = getattr(self._bins, 'bin_centers', None)
            if self._bin_centers is None:
                self._bin_centers = _bin_edges_to_bin_centers(self.bin_edges)
        return self._bin_centers

    @property
    def labels(self):
        if self._labels is None:
            self._labels = getattr(self._bins, 'labels', None)
        return self._labels

    @labels.setter
    def labels(self, value):
        if value is None:
            self._labels = None
        elif isinstance(value, str):
            self._labels = [value]
        elif isinstance(value, (tuple, list)):
            self._labels = list(value)
        else:
            raise TypeError("`labels` must be str or list of str or None.")

        if self.labels is not None and len(self.labels) != self.dimension:
            self._labels = None
            raise ValueError("`labels` must have a length of `dimension`.")

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self._bins, attr)

    @property
    def is_equally_sized(self) -> np.array:
        """True for each dimension if all bins are of the same size."""
        if _is_1d_array_of_scalar(self.bin_size):
            return tuple(True for _ in self.bin_size)
        elif _is_2d_array_of_1d_array_of_scalar(self.bin_size):
            return tuple(all(np.isclose(bs, bs[0])) for bs in self.bin_size)

        elif _is_2d_inhomogeneous_array(self.bin_size):
            result = []
            for bs in self.bin_size:
                if _is_scalar(bs):
                    result.append(True)
                else:
                    result.append(all(np.isclose(bs, bs[0])))
            return tuple(result)
        else:
            raise TypeError

    def equalize_bin_size(self):
        """
        Return a new instance of `Bins` with bin_size set equal to the first bin_size element in each dimension
        and extend_range=None.
        """
        new_bin_size = []
        for bs in self.bin_size:
            try:
                new_bin_size.append(bs[0])
            except IndexError:
                new_bin_size.append(bs)
        return Bins(bin_size=new_bin_size, bin_range=self.bin_range, extend_range=None)


def _histogram_fast_histogram(data, bins) -> np.ndarray:
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data : array-like
        Coordinate values with shape (dimensions, n_points) to be binned
    bins : Bins
        The bin specification

    Returns
    -------
    numpy.ndarray
    """
    if data.shape[0] == 1:
        img = fast_histogram.histogram1d(data, range=bins.bin_range[0], bins=bins.n_bins[0])
    elif data.shape[0] == 2:
        img = fast_histogram.histogram2d(*data, range=bins.bin_range, bins=bins.n_bins)
    else:
        raise TypeError('Dimension of data must be 1 or 2.')
    return img


def _histogram_boost_histogram(data, bins) -> np.ndarray:
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data : array-like
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins : Bins
        The bin specification

    Returns
    -------
    numpy.ndarray
    """
    hist_axes = [bh.axis.Variable(bins.bin_edges[i], metadata=bins.labels[i]) for i in range(bins.dimension)]
    hist = bh.Histogram(*hist_axes).fill(*data)
    img = hist.view()
    return img


def _histogram_mean_fast_histogram(data, bins, values):
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data : array-like
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins : Bins
        The bin specification
    values : array-like
        Values with shape (n_points,) to be averaged in each bin

    Returns
    -------
    numpy.ndarray
    """
    if data.shape[0] == 1:
        hist = fast_histogram.histogram1d(data, range=bins.bin_range[0], bins=bins.n_bins[0])
        hist_w = fast_histogram.histogram1d(data, range=bins.bin_range[0], bins=bins.n_bins[0], weights=values)
    elif data.shape[0] == 2:
        hist = fast_histogram.histogram2d(*data, range=bins.bin_range, bins=bins.n_bins)
        hist_w = fast_histogram.histogram2d(*data, range=bins.bin_range, bins=bins.n_bins, weights=values)
    else:
        raise TypeError('Dimension of data must be 1 or 2.')

    with np.errstate(divide='ignore', invalid='ignore'):
        hist = np.true_divide(hist_w, hist)
        hist[hist == np.inf] = 0
        # hist = np.nan_to_num(hist)
    return hist


def _histogram_mean_boost_histogram(data, bins, values) -> np.ndarray:
    """
    Provide histogram with averaged values for all counts in each bin.

    Parameters
    ----------
    data : array-like
        Coordinate values with shape (n_dimensions, n_points) to be binned
    bins : Bins
        The bin specification

    Returns
    -------
    numpy.ndarray
    """
    hist_axes = [bh.axis.Variable(bins.bin_edges[i], metadata=bins.labels[i]) for i in range(bins.dimension)]
    hist = bh.Histogram(*hist_axes).fill(*data)
    hist_w = bh.Histogram(*hist_axes).fill(*data, weight=values)
    hist = hist_w / hist
    return hist.view()


def _check_loc_properties(locdata:LocData, loc_properties:Union[str, Iterable]):
    """
    Check that loc_properties are valid properties in locdata.

    Parameters
    ----------
    locdata
        Localization data
    loc_properties
        LocData property names

    Returns
    -------
    list[str]
        Valid localization property names
    """
    if loc_properties is None:  # use coordinate_labels
        labels = locdata.coordinate_labels.copy()
    elif isinstance(loc_properties, str):
        if loc_properties not in locdata.data.columns:
            raise ValueError(f'{loc_properties} is not a valid property in locdata.data.')
        labels = [loc_properties]
    elif isinstance(loc_properties, (tuple, list)):
        labels = list(loc_properties)
        for loc_property in loc_properties:
            if loc_property not in locdata.data.columns:
                raise ValueError(f'{loc_property} is not a valid property in locdata.data.')
    else:
        raise ValueError(f'{loc_properties} is not a valid property in locdata.data.')
    return labels


def histogram(locdata, loc_properties=None, other_property=None,
              bins=None, n_bins=None, bin_size=None, bin_edges=None, bin_range=None,
              ):
    """
    Make histogram of loc_properties (columns in locdata.data) by binning all localizations
    or averaging other_property within each bin.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list of str or None
        Localization properties to be grouped into bins. If None The coordinate_values of locdata are used.
    other_property : str or None
        Localization property that is averaged in each pixel. If None localization counts are
        shown.
    bins : int or sequence or `Bins` or `boost_histogram.axis.Axis` or None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray of float with shape (dimension, n_bin_edges) or None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple or numpy.ndarray or None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple or numpy.ndarray or None
        The size of bins in units of locdata coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple or tuple of tuples of float with shape (dimension, 2) or None or 'zero'
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.

    Returns
    -------
    namedtuple('Histogram', "data bins labels"): (numpy.ndarray, `Bins`, list)
    """
    labels_ = _check_loc_properties(locdata, loc_properties)
    data = locdata.data[labels_].values.T

    if (bin_range is None or isinstance(bin_range, str)) and bin_edges is None:
        bin_range_ = ranges(locdata, loc_properties=labels_, special=bin_range)
    else:
        bin_range_ = bin_range

    try:
        bins = Bins(bins, n_bins, bin_size, bin_edges, bin_range_, labels=labels_)
    except ValueError:
        raise ValueError("Bin dimension and len of `loc_properties` is incompatible.")

    if other_property is None:
        # histogram data by counting points
        if data.shape[0] == 2:
            # we are using fast-histogram for 2D since it is even faster than boost_histogram
            img = _histogram_fast_histogram(data, bins)
        elif data.shape[0] == 1 or data.shape[0] == 3:
            img = _histogram_boost_histogram(data, bins)
        else:
            raise TypeError('loc_properties must contain a string or a list with 1, 2 or 3 elements.')
        labels_.append('counts')

    elif other_property in locdata.data.columns:
        # histogram data by averaging values
        values = locdata.data[other_property].values
        if data.shape[0] == 2:
            img = _histogram_mean_fast_histogram(data, bins, values)
        elif data.shape[0] == 1 or data.shape[0] == 3:
            img = _histogram_mean_boost_histogram(data, bins, values)
        else:
            raise TypeError('No more than 3 elements in loc_properties are allowed.')
        labels_.append(other_property)
    else:
        raise TypeError(f'Parameter for `other_property` {other_property} is not a valid property name.')

    Histogram = namedtuple('Histogram', "data bins labels")
    return Histogram(img, bins, labels_)



"""
Standard analysis procedures.

This module contains classes and functions for carrying out standardized analysis procedures on localization data.
All functions typically take LocData objects as input and provide an analysis class with derived analysis results
and standard functions for presentation.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   localization_precision
   localization_property
   localization_property_2d
   localization_property_correlations
   localizations_per_frame
   nearest_neighbor
   pipeline
   ripley
   uncertainty
   cbc
   accumulation_analysis
   blinking
   drift

"""

# from locan.analysis.analysis_example import AnalysisExampleAlgorithm_1, AnalysisExampleAlgorithm_2
from locan.analysis.localization_precision import *
from locan.analysis.localization_property import *
from locan.analysis.localization_property_2d import *
from locan.analysis.localization_property_correlations import *
from locan.analysis.localizations_per_frame import *
from locan.analysis.nearest_neighbor import *
from locan.analysis.pipeline import *
from locan.analysis.ripley import *
from locan.analysis.uncertainty import *
from locan.analysis.cbc import *
from locan.analysis.accumulation_analysis import *
from locan.analysis.blinking import *
from locan.analysis.drift import *

__all__ = []
__all__.extend(localization_precision.__all__)
__all__.extend(localization_property.__all__)
__all__.extend(localization_property_2d.__all__)
__all__.extend(localization_property_correlations.__all__)
__all__.extend(localizations_per_frame.__all__)
__all__.extend(nearest_neighbor.__all__)
__all__.extend(pipeline.__all__)
__all__.extend(ripley.__all__)
__all__.extend(uncertainty.__all__)
__all__.extend(cbc.__all__)
__all__.extend(accumulation_analysis.__all__)
__all__.extend(blinking.__all__)
__all__.extend(drift.__all__)



from copy import copy

import google.protobuf.message
import pytest

from locan import metadata_to_formatted_string, metadata_from_toml, message_scheme
from locan.data import metadata_pb2
from locan.data.metadata_utils import _modify_meta, _dict_to_protobuf


@pytest.fixture
def metadata():
    metadata = metadata_pb2.Metadata()
    optical_unit = metadata.experiment.setups.add().optical_units.add()
    optical_unit.detection.camera.integration_time.FromMilliseconds(20)

    metadata.identifier = "111"
    metadata.comment = "comment"
    metadata.creation_time.FromJsonString("1111-11-11T11:11:11+01:00")

    return metadata


@pytest.fixture
def meta_dict():
    meta_dict_ = {'identifier': '123',
                 'comment': 'my comment',
                 'ancestor_identifiers': ['1', '2'],
                 'production_time': '2022-05-14T06:58:00Z',
                 'localizer': {'software': 'rapidSTORM'},
                 'relations': [{'identifier': '1'}],
                 'experiment': {'setups': [{'identifier': '1',
                    'optical_units': [{'identifier': '1',
                      'detection': {'camera': {'identifier': '1',
                        'integration_time': 10000000}}}]}]}}
    return meta_dict_


@pytest.fixture
def metadata_toml():
    metadata_toml_ = \
    """
    # Define the class (message) instances.
    
    [[messages]]
    name = "metadata"
    module = "locan.data.metadata_pb2"
    class_name = "Metadata"
    
    
    # Fill metadata attributes
    # Use [[]] to add repeated elements
    # Use string '2022-05-14T06:58:00Z' for Timestamp elements
    # Use int in nanoseconds for Duration elements
    
    [metadata]
    identifier = "123"
    comment = "my comment"
    ancestor_identifiers = ["1", "2"]
    production_time = '2022-05-14T06:58:00Z'
    
    [metadata.localizer]
    software = "rapidSTORM"
    
    [[metadata.relations]]
    identifier = "1"
    
    [[metadata.experiment.setups]]
    identifier = "1"
    
    [[metadata.experiment.setups.optical_units]]
    identifier = "1"
    
    [metadata.experiment.setups.optical_units.detection.camera]
    identifier = "1"
    integration_time = 10_000_000
    """
    return metadata_toml_


def test__modify_meta(metadata, locdata_2d):
    new_locdata = copy(locdata_2d)
    new_metadata = _modify_meta(locdata_2d, new_locdata,
                                function_name="test_function",
                                parameter=dict(first=1, second=2),
                                meta=None)

    assert metadata_to_formatted_string(message=new_metadata.history[-1], as_one_line=True) == \
           "name: \"test_function\" parameter: \"{\\\'first\\\': 1, \\\'second\\\': 2}\""


def test_metadata_to_formatted_string(metadata):
    results_string = metadata_to_formatted_string(message=metadata, as_one_line=True)
    results_string_expected = 'identifier: "111" comment: "comment" experiment { setups { optical_units { detection ' \
                              '{ camera { integration_time { 0.020s } } } } } } creation_time { 1111-11-11T10:11:11Z }'
    assert results_string == results_string_expected


def test__dict_to_protobuf(meta_dict):
    metadata = metadata_pb2.Metadata()
    metadata.comment = "to be changed"
    metadata.element_count = 1  # will not be touched
    new_metadata = _dict_to_protobuf(dictionary=meta_dict, message=metadata)
    assert isinstance(new_metadata, google.protobuf.message.Message)
    results_string = metadata_to_formatted_string(message=new_metadata, as_one_line=True)
    results_string_expected = 'identifier: "123" comment: "my comment" ancestor_identifiers: "1" ' \
                              'ancestor_identifiers: "2" element_count: 1 relations { identifier: "1" } ' \
                              'experiment { setups { identifier: "1" ' \
                              'optical_units { identifier: "1" detection { ' \
                              'camera { identifier: "1" integration_time { 0.010s } } } } } } ' \
                              'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    assert results_string == results_string_expected

    metadata = metadata_pb2.Metadata()
    metadata.comment = "to be changed"
    metadata.element_count = 1  # will not be touched
    new_metadata = _dict_to_protobuf(dictionary=meta_dict, message=metadata, inplace=True)
    assert new_metadata is None
    results_string = metadata_to_formatted_string(message=metadata, as_one_line=True)
    results_string_expected = 'identifier: "123" comment: "my comment" ancestor_identifiers: "1" ' \
                              'ancestor_identifiers: "2" element_count: 1 relations { identifier: "1" } ' \
                              'experiment { setups { identifier: "1" ' \
                              'optical_units { identifier: "1" detection { ' \
                              'camera { identifier: "1" integration_time { 0.010s } } } } } } ' \
                              'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    assert results_string == results_string_expected


def test_metadata_from_toml(metadata_toml):
    result = metadata_from_toml(metadata_toml)
    assert isinstance(result, dict)
    assert isinstance(result['metadata'], google.protobuf.message.Message)
    results_string = metadata_to_formatted_string(message=result['metadata'], as_one_line=True)
    results_string_expected = 'identifier: "123" comment: "my comment" ancestor_identifiers: "1" ' \
                              'ancestor_identifiers: "2" relations { identifier: "1" } experiment { ' \
                              'setups { identifier: "1" optical_units { identifier: "1" detection { ' \
                              'camera { identifier: "1" integration_time { 0.010s } } } } } } ' \
                              'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    assert results_string == results_string_expected


def test_message_scheme(metadata):
    scheme = message_scheme(metadata)
    assert isinstance(scheme, dict)
    assert scheme['creation_time'] == '1111-11-11T10:11:11Z'
    assert scheme['modification_time'] == '1970-01-01T00:00:00Z'



"""

Constants used throughout the project.

.. autosummary::
   :toctree: ./

   PROPERTY_KEYS
   RAPIDSTORM_KEYS
   ELYRA_KEYS
   THUNDERSTORM_KEYS
   NANOIMAGER_KEYS
   SMLM_KEYS
   DECODE_KEYS
   SMAP_KEYS
"""
from __future__ import annotations

from enum import Enum, unique
from dataclasses import dataclass

from locan.dependencies import HAS_DEPENDENCY
if HAS_DEPENDENCY["colorcet"]: from colorcet import m_fire, m_gray, m_coolwarm, m_glasbey_dark

import matplotlib.colors as mcolors


__all__ = ['PROPERTY_KEYS', 'PropertyKey', 'PropertyDescription',
           'HullType', 'FileType', 'RenderEngine', 'ColorMaps',
           'RAPIDSTORM_KEYS', 'ELYRA_KEYS', 'THUNDERSTORM_KEYS', 'NANOIMAGER_KEYS', 'SMLM_KEYS', 'DECODE_KEYS',
           'SMAP_KEYS'
           ]


@dataclass
class PropertyDescription:
    """
    A property of a localization or group of localizations
    representing column names in `LocData.data` and `LocData.properties`.

    Attributes
    ----------
    name:
        property name
    type:
        property type
    unit_SI:
        SI unit that is appropriate for property
    unit:
        The actual unit currently used
    description:
        Explanation what the property represents
    """
    name: str
    type: str
    unit_SI: str | None = None
    unit: str | None = None
    description: str = ""


@unique
class PropertyKey(Enum):
    """
    Property descriptions for standard properties used in `locan.LocData` and throughout locan.
    """
    index = PropertyDescription(
        'index',
        'integer',
        description="a positive integer identifying the localization")
    original_index = PropertyDescription(
        'original_index',
        'integer')
    position_x = PropertyDescription(
        'position_x',
        'float',
        unit_SI='m',
        description="spatial coordinate.")
    position_y = PropertyDescription(
        'position_y',
        'float',
        unit_SI='m',
        description="spatial coordinate.")
    position_z = PropertyDescription(
        'position_z',
        'float',
        unit_SI='m',
        description="spatial coordinate.")
    frame = PropertyDescription(
        'frame',
        'integer',
        description="frame  number in which the localization occurs")
    frames_number = PropertyDescription(
        'frames_number',
        'integer',
        description="number of frames that contribute to a merged localization")
    frames_missing = PropertyDescription(
        'frames_missing',
        'integer',
        description="number of frames that occurred between two successive localizations")
    time = PropertyDescription(
        'time',
        'float',
        unit_SI="s")
    intensity = PropertyDescription(
        'intensity',
        'float',
        description="intensity or emission strength as estimated by the fitter")
    local_background = PropertyDescription(
        'local_background',
        'float',
        description="background in the neighborhood of localization as estimated by the fitter")
    local_background_sigma = PropertyDescription(
        'local_background_sigma',
        'float',
        description="variation of local background in terms of standard deviation")
    signal_noise_ratio = PropertyDescription(
        'signal_noise_ratio',
        'float',
        description="ratio between mean intensity (i.e. intensity for a single localization) "
                    "and the standard deviation of local_background "
                    "(i.e. local_background_sigma for a single localization)")
    signal_background_ratio = PropertyDescription(
        'signal_background_ratio',
        'float',
        description="ratio between mean intensity (i.e. intensity for a single localization) and the local_background")
    chi_square = PropertyDescription(
        'chi_square',
        'float',
        description="chi-square value of the fitting procedure as estimated by the fitter")
    two_kernel_improvement = PropertyDescription(
        'two_kernel_improvement',
        'float',
        description="a rapidSTORM parameter describing the improvement from two kernel fitting")
    psf_amplitude = PropertyDescription('psf_amplitude', 'float')
    psf_width = PropertyDescription(
        'psf_width',
        'float',
        description="full-width-half-max of the fitted Gauss-function - "
                    "being isotropic or representing the root-mean-square of psf_width_c for all dimensions")
    psf_width_x = PropertyDescription(
        'psf_width_x',
        'float',
        description="full-width-half-max of the fitted Gauss-function in x-dimension as estimated by the fitter")
    psf_width_y = PropertyDescription(
        'psf_width_y',
        'float',
        description="full-width-half-max of the fitted Gauss-function in y-dimension as estimated by the fitter")
    psf_width_z = PropertyDescription(
        'psf_width_z',
        'float',
        description="full-width-half-max of the fitted Gauss-function in z-dimension as estimated by the fitter")
    psf_half_width = PropertyDescription('psf_half_width', 'float')
    psf_half_width_x = PropertyDescription('psf_half_width_x', 'float')
    psf_half_width_y = PropertyDescription('psf_half_width_y', 'float')
    psf_half_width_z = PropertyDescription('psf_half_width_z', 'float')
    psf_sigma = PropertyDescription(
        'psf_sigma',
        'float',
        description="sigma of the fitted Gauss-function - "
                    "being isotropic or representing the root-mean-square of psf_sigma_c for all dimensions")
    psf_sigma_x = PropertyDescription(
        'psf_sigma_x',
        'float',
        description="sigma of the fitted Gauss-function in x-dimension as estimated by the fitter")
    psf_sigma_y = PropertyDescription(
        'psf_sigma_y',
        'float',
        description="sigma of the fitted Gauss-function in y-dimension as estimated by the fitter")
    psf_sigma_z = PropertyDescription(
        'psf_sigma_z',
        'float',
        description="sigma of the fitted Gauss-function in z-dimension as estimated by the fitter")
    uncertainty = PropertyDescription(
        'uncertainty',
        'float',
        description="localization error for all dimensions or representing a value proportional to "
                    "psf_sigma / sqrt(intensity) or "
                    "representing the root-mean-square of uncertainty_c for all dimensions.")
    uncertainty_x = PropertyDescription(
        'uncertainty_x',
        'float',
        description="localization error in x-dimension estimated by a fitter "
                    "or representing a value proportional to psf_sigma_x / sqrt(intensity)")
    uncertainty_y = PropertyDescription(
        'uncertainty_y',
        'float',
        description="localization error in y-dimension estimated by a fitter "
                    "or representing a value proportional to psf_sigma_y / sqrt(intensity)")
    uncertainty_z = PropertyDescription(
        'uncertainty_z',
        'float',
        description="localization error in z-dimension estimated by a fitter "
                    "or representing a value proportional to psf_sigma_z / sqrt(intensity)")
    channel = PropertyDescription(
        'channel',
        'integer',
        description="identifier for an imaging channel")
    slice_z = PropertyDescription(
        'slice_z',
        'float',
        description="identifier for an imaging slice")
    plane = PropertyDescription(
        'plane',
        'integer',
        description="identifier for an imaging plane")
    cluster_label = PropertyDescription(
        'cluster_label',
        'integer',
        description="identifier for a localization cluster")

    @classmethod
    def coordinate_labels(cls) -> list[PropertyKey]:
        """Property descriptions for properties representing spatial coordinates"""
        return [cls.position_x, cls.position_y, cls.position_z]

    @classmethod
    def summary(cls) -> str:
        """A formatted string representation of PropertyKey showing all elements with attributes."""
        lines = []
        for element in cls:
            lines.append(f"{element.value.name}")
            lines.append(f"type: {element.value.type}, unit_SI: {element.value.unit_SI}, unit: {element.value.unit}")
            lines.append(f"{element.value.description}\n")
        return '\n'.join(lines)


#: Keys for the most common LocData properties.
#: Values suggest a type for conversion.
#: If 'integer', 'signed', 'unsigned', 'float' :func:`pandas.to_numeric` can be applied.
#: Otherwise :func:`pandas.astype` can be applied.
PROPERTY_KEYS = {item.name: item.value.type for item in PropertyKey if item.value.type is not None}


class HullType(Enum):
    """
    Hull definitions that are supported for `LocData` objects.
    """
    BOUNDING_BOX = 'bounding_box'
    CONVEX_HULL = 'convex_hull'
    ORIENTED_BOUNDING_BOX = 'oriented_bounding_box'
    ALPHA_SHAPE = 'alpha_shape'


# File types
class FileType(Enum):
    """
    File types for localization files.

    The listed file types are supported with input/output functions in :func:`io.io_locdata`.
    The types correspond to the metadata keys for LocData objects. That is they are equal to the file types in
    the protobuf message `locan.data.metadata_pb2.Metadata`.
    """
    UNKNOWN_FILE_TYPE = 0
    CUSTOM = 1
    RAPIDSTORM = 2
    ELYRA = 3
    THUNDERSTORM = 4
    ASDF = 5
    NANOIMAGER = 6
    RAPIDSTORMTRACK = 7
    SMLM = 8
    DECODE = 9
    SMAP = 10


class RenderEngine(Enum):
    """
    Engine to be use for rendering and displaying localization data as 2d or 3d images.

    Each engine represents a library to be used as backend for rendering and plotting.
    """
    if not HAS_DEPENDENCY["mpl_scatter_density"]:
        _ignore_ = 'MPL_SCATTER_DENSITY'
    if not HAS_DEPENDENCY["napari"]:
        _ignore_ = 'NAPARI'
    MPL = 0
    """matplotlib"""
    MPL_SCATTER_DENSITY = 1
    """mpl-scatter-density"""
    NAPARI = 2
    """napari"""


class ColorMaps(Enum):
    """
    Preferred colormaps to be used for visualization.
    """
    if HAS_DEPENDENCY["colorcet"]:
        COLORMAP_CONTINUOUS = m_fire
        COLORMAP_CONTINUOUS_GRAY = m_gray
        COLORMAP_DIVERGING = m_coolwarm
        COLORMAP_CATEGORICAL = m_glasbey_dark
    else:
        COLORMAP_CONTINUOUS = mcolors.Colormap('viridis')
        COLORMAP_DIVERGING = mcolors.Colormap('coolwarm')
        COLORMAP_CATEGORICAL = mcolors.Colormap('tab20')


#: Mapping column names in RapidSTORM files to `LocData` property keys
RAPIDSTORM_KEYS = {
    'Position-0-0': 'position_x',
    'Position-1-0': 'position_y',
    'Position-2-0': 'position_z',
    'ImageNumber-0-0': 'frame',
    'Amplitude-0-0': 'intensity',
    'FitResidues-0-0': 'chi_square',
    'LocalBackground-0-0': 'local_background',
    'TwoKernelImprovement-0-0': 'two_kernel_improvement',
    'Position-0-0-uncertainty': 'uncertainty_x',
    'Position-1-0-uncertainty': 'uncertainty_y',
    'Position-2-0-uncertainty': 'uncertainty_z'
}


#: Mapping column names in Zeiss Elyra files to `LocData` property keys
ELYRA_KEYS = {
    'Index': 'original_index',
    'First Frame': 'frame',
    'Number Frames': 'frames_number',
    'Frames Missing': 'frames_missing',
    'Position X [nm]': 'position_x',
    'Position Y [nm]': 'position_y',
    'Position Z [nm]': 'position_z',
    'Precision [nm]': 'uncertainty',
    'Number Photons': 'intensity',
    'Background variance': 'local_background_sigma',
    'Chi square': 'chi_square',
    'PSF half width [nm]': 'psf_half_width',
    'PSF width [nm]': 'psf_width',
    'Channel': 'channel',
    'Z Slice': 'slice_z'
}


#: Mapping column names in Thunderstorm files to `LocData` property keys
THUNDERSTORM_KEYS = {
    'id': 'original_index',
    'frame': 'frame',
    'x [nm]': 'position_x',
    'y [nm]': 'position_y',
    'z [nm]': 'position_z',
    'uncertainty [nm]': 'uncertainty',
    'uncertainty_xy [nm]': 'uncertainty_x',
    'uncertainty_z [nm]': 'uncertainty_z',
    'intensity [photon]': 'intensity',
    'offset [photon]': 'local_background',
    'bkgstd [photon]': 'local_background_sigma',
    'chi2': 'chi_square',
    'sigma1 [nm]': 'psf_sigma_x',
    'sigma2 [nm]': 'psf_sigma_y',
    'sigma [nm]': 'psf_sigma',
    'detections': 'frames_number'
}


#: Mapping column names in Nanoimager files to `LocData` property keys
NANOIMAGER_KEYS = {
    'Channel': 'channel',
    'Frame': 'frame',
    'X (nm)': 'position_x',
    'Y (nm)': 'position_y',
    'Z (nm)': 'position_z',
    'Photons': 'intensity',
    'Background': 'local_background'
}


#: Mapping column names in SMLM files to `LocData` property keys
SMLM_KEYS = {
    'id': 'original_index',
    'frame': 'frame',
    'x': 'position_x',
    'y': 'position_y',
    'z': 'position_z',
    'x_position': 'position_x',
    'y_position': 'position_y',
    'z_position': 'position_z',

    'uncertainty [nm]': 'uncertainty',
    'uncertainty_xy [nm]': 'uncertainty_x',
    'uncertainty_z [nm]': 'uncertainty_z',
    'intensity': 'intensity',
    'Amplitude_0_0': 'intensity',
    'background': 'local_background',
    'LocalBackground_0_0': 'local_background',
    'FitResidues_0_0': 'chi_square',
}

#: Mapping column names in DECODE files to `LocData` property keys
DECODE_KEYS = {
    'id': 'original_index',
    'frame_ix': 'frame',
    'x': 'position_x',
    'y': 'position_y',
    'z': 'position_z',
    'bg': 'local_background',
    'phot': 'intensity',
}

#: Mapping column names in SMAP files to `LocData` property keys
SMAP_KEYS = {
    'frame': 'frame',
    'xnm': 'position_x',
    'ynm': 'position_y',
    'znm': 'position_z',
    'bg': 'local_background',
    'phot': 'intensity',
    'channel': 'channel',
    'xnmerr': 'uncertainty_x',
    'ynmerr': 'uncertainty_y',
    'znmerr': 'uncertainty_z',
}



import os
from pathlib import Path

import tifffile as tif
import pytest

from locan.dependencies import HAS_DEPENDENCY
from locan import ROOT_DIR
from locan import load_locdata
from locan.scripts.script_check import render_locs_per_frame_napari, sc_check

if HAS_DEPENDENCY["napari"]: import napari


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_locs_per_frame_napari(locdata_2d):
    images_path = Path(ROOT_DIR / 'tests/test_data/images.tif')
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / 'tests/test_data/rapidSTORM_from_images.txt')
    assert locdata_path.exists()

    image_stack = tif.imread(str(images_path))
    locdata = load_locdata(locdata_path, file_type=2)

    viewer = render_locs_per_frame_napari(images=image_stack, pixel_size=133, locdata=locdata,
                                          viewer=None, transpose=True,
                                          kwargs_image=None, kwargs_points=None)
    assert viewer
    # napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_check():
    images_path = Path(ROOT_DIR / 'tests/test_data/images.tif')
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / 'tests/test_data/rapidSTORM_from_images.txt')
    assert locdata_path.exists()
    sc_check(pixel_size=133, file_images=images_path, file_locdata=locdata_path, file_type=2)


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_check_from_sys(capfd):
    images_path = Path(ROOT_DIR / 'tests/test_data/images.tif')
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / 'tests/test_data/rapidSTORM_from_images.txt')
    assert locdata_path.exists()
    exit_status = os.system(f'locan check 133 -f {str(images_path)} -l {str(locdata_path)} -t 2')
    captured = capfd.readouterr()
    print(captured.out)
    assert exit_status == 0



"""

Filter localization data.

This module provides functions for filtering LocData objects.
The functions take LocData as input and compute new LocData objects.

"""
import sys

import numpy as np
from sklearn.neighbors import NearestNeighbors

from locan.data.locdata import LocData
from locan.constants import HullType
from locan.configuration import N_JOBS
from locan.data.metadata_utils import _modify_meta
from locan.data.rois import Roi
from locan.data.region import Region, Region2D, RoiRegion


__all__ = ['select_by_condition', 'select_by_region', 'select_by_image_mask', 'exclude_sparse_points',
           'random_subset', 'localizations_in_cluster_regions']


def select_by_condition(locdata, condition):
    """
    Select by specifying conditions on data properties.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to select localization data.
    condition : string
        Conditions as input in select method.
        More precise: query specifications to be used with pandas query.

    Returns
    -------
    LocData
        a new instance of LocData referring to the specified dataset.
    """
    local_parameter = locals()

    # select
    new_indices = locdata.data.query(condition).index.values.tolist()

    # instantiate
    new_locdata = LocData.from_selection(locdata=locdata, indices=new_indices)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter, meta=None)
    new_locdata.meta = meta_

    return new_locdata


def select_by_region(locdata, region, loc_properties=None, reduce=True):
    """
    Select localizations from `locdata` that are within `region` and return a new LocData object.
    Selection is with respect to loc_properties or to localization coordinates that correspond to region dimension.

    Parameters
    ----------
    locdata : LocData
        Localization data that is tested for being inside the region.
    region : Region
        Tested region
    loc_properties : list of string or None
        Localization properties to be tested.
    reduce : Bool
        Return the reduced LocData object or keep references alive.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations within region of interest.

    Note
    ----
    Points on boundary of regions are considered outside if region is a shapely object but inside if region is a
    matplotlib (which is the case for RoiRegion) object.
    """
    local_parameter = locals()

    if loc_properties is None:
        loc_properties_ = locdata.coordinate_labels[0:region.dimension]
    else:
        loc_properties_ = loc_properties

    points = locdata.data[list(loc_properties_)].values

    if isinstance(region, (Region2D, RoiRegion)):
        indices_inside = region.contains(points)
        locdata_indices_to_keep = locdata.data.index[indices_inside]
        new_locdata = LocData.from_selection(locdata=locdata, indices=locdata_indices_to_keep)
        new_locdata.region = region
    else:
        raise NotImplementedError("Only Region2D has been implemented.")

    # finish
    if reduce:
        new_locdata.reduce()

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter,
                         meta=None)
    new_locdata.meta = meta_

    return new_locdata


def select_by_image_mask(selection, mask, pixel_size):
    """
    Select by masking using a binary image(e.g. generated by thresholding a transmitted-light microscopy image.

    Parameters
    ----------
    selection : Selection
        specifying the localization data from which to select localization data.
    mask :
        binary image.
    pixel_size : tuple(float)
        pixel sizes for each dimension in units of localization coordinates.

    Returns
    -------
    Selection
        a new instance of Selection referring to the specified dataset.
    """
    raise NotImplementedError


def exclude_sparse_points(locdata, other_locdata=None, radius=50, min_samples=5):
    """
    Exclude localizations by thresholding a local density.

    A subset of localizations, that exhibit a small local density of localizations from locdata or alternatively from
    other_locdata, is identified as noise and excluded.
    Noise is identified by using a nearest-neighbor search (:class:`sklearn.neighbors.NearestNeighbors`) to find all
    localizations within a circle (sphere) of the given `radius`. If the number of localizations is below the
    threshold value `min_samples`, the localization is considered to be noise.

    The method identifies the same noise points as done by the clustering algorithm DBSCAN [1]_.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to exclude localization data.
    other_locdata : LocData
        Specifying the localization data on which to compute local density.
    radius: float
        Radius of a circle or sphere in which neighbors are identified (equivalent to epsilon in DBSCAN).
    min_samples : int
        The minimum number of samples in the neighborhood that need to be found for each localization to not be
        identified as noise (equivalent to minPoints in DBSCAN).

    Returns
    -------
    Tuple of LocData
        Two Locdata objects are returned, one carrying all noise localizations, the other carrying all localizations
        except noise.

    References
    ----------
    .. [1] Martin Ester, Hans-Peter Kriegel, Jörg Sander, Xiaowei Xu,
       A density-based algorithm for discovering clusters in large spatial databases with noise.
       In: Evangelos Simoudis, Jiawei Han, Usama M. Fayyad (Hrsg.): Proceedings of the Second International Conference
       on Knowledge Discovery and Data Mining (KDD-96). AAAI Press, 1996, S. 226-231, ISBN 1-57735-004-9.
    """
    local_parameter = locals()

    if other_locdata is None:
        nn = NearestNeighbors(metric='euclidean', n_jobs=N_JOBS).fit(locdata.coordinates)
        neighbor_points_list = nn.radius_neighbors(radius=radius, return_distance=False)
        # if points is not provided the query point is not considered its own neighbor.
    else:
        nn = NearestNeighbors(metric='euclidean', n_jobs=N_JOBS).fit(other_locdata.coordinates)
        neighbor_points_list = nn.radius_neighbors(locdata.coordinates, radius=radius, return_distance=False)

    indices_to_keep = [len(pts) >= min_samples for pts in neighbor_points_list]
    locdata_indices_to_keep = locdata.data.index[indices_to_keep]
    new_locdata = LocData.from_selection(locdata, locdata_indices_to_keep)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter, meta=None)
    new_locdata.meta = meta_

    return new_locdata


def random_subset(locdata, n_points, replace=True, seed=None):
    """
    Take a random subset of localizations.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to select localization data.
    n_points : int
        Number of localizations to randomly choose from locdata.
    replace : bool
        Indicate if sampling is with or without replacement
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        Random number generation seed

    Returns
    -------
    LocData
        a new instance of LocData carrying the subset of localizations.
    """
    local_parameter = locals()

    if not locdata:
        return locdata

    rng = np.random.default_rng(seed)

    indices = rng.choice(locdata.data.index, size=n_points, replace=replace)
    new_locdata = LocData.from_selection(locdata, indices)

    # update metadata
    meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                         parameter=local_parameter, meta=None)
    new_locdata = LocData.from_selection(locdata, indices, meta=meta_)

    return new_locdata


def localizations_in_cluster_regions(locdata, collection, hull_type=HullType.CONVEX_HULL):
    """
    Identify localizations from `locdata` within the regions of all `collection` elements.

    Parameters
    ----------
    locdata : LocData
        Localization data that is tested for being inside the region

    collection : LocData or list(LocData)
        A set of Locdata objects collected in a collection or list.

    hull_type : HullType
        The hull type for each LocData object that is used to define the region.

    Returns
    --------
    LocData
        A collection of LocData objects with all elements of locdata contained by the region.
    """
    locdatas = []
    if isinstance(collection, LocData):
        if isinstance(collection.references, list):  # this case covers pure collections
            for ref in collection.references:
                cregion = getattr(ref, hull_type.value).region
                locdata_selection = select_by_region(locdata=locdata, region=cregion)
                locdatas.append(locdata_selection)
        else:  # this case covers selections of collections
            for index in collection.indices:
                cregion = getattr(collection.references.references[index], hull_type.value).region
                locdata_selection = select_by_region(locdata=locdata, region=cregion)
                locdatas.append(locdata_selection)
    else:  # this case covers list of LocData objects
        for ref in collection:
            cregion = getattr(ref, hull_type.value).region
            locdata_selection = select_by_region(locdata=locdata, region=cregion)
            locdatas.append(locdata_selection)

    new_collection = LocData.from_collection(locdatas)

    return new_collection



import copy

import pytest
import numpy as np
import pandas as pd

from locan.data import metadata_pb2
from locan import LocData
from locan import Region, Rectangle
from locan import AlphaShape


# fixtures for DataFrames (fixtures for LocData are defined in conftest.py)

@pytest.fixture()
def df_simple():
    dict_ = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1]
    }
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_line():
    dict_ = {
        'position_x': [1, 2, 3, 4, 5],
        'position_y': [1, 2, 3, 4, 5]
    }
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_empty():
    dict_ = {
    }
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_other_simple():
    dict_ = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [10, 11, 13, 14, 11]
        }
    return pd.DataFrame.from_dict(dict_)


# tests

COMMENT_METADATA = metadata_pb2.Metadata(comment='some user comment')


def test_LocData(df_simple, caplog):
    dat = LocData(dataframe=df_simple, meta=COMMENT_METADATA)
    assert len(dat) == 5
    assert dat.coordinate_labels == ['position_x', 'position_y']
    assert dat.dimension == 2
    assert np.array_equal(dat.centroid, [2., 1.8])
    assert dat.centroid[0] == dat.properties['position_x']
    for x, y in zip(dat.coordinates, [[0, 0], [0, 1], [1, 3], [4, 4], [5, 1]]):
        assert np.all(x == np.array(y))
    assert dat.meta.comment == COMMENT_METADATA.comment
    # assert dat.meta.identifier == '1'  # the test runs ok when testing this function alone.
    assert dat.bounding_box.region_measure == 20
    assert 'region_measure_bb' in dat.properties
    assert 'localization_density_bb' in dat.properties
    assert dat.convex_hull.region_measure == 12.5
    assert 'region_measure_ch' in dat.properties
    assert 'localization_density_ch' in dat.properties
    assert round(dat.oriented_bounding_box.region_measure) == 16
    assert 'region_measure_obb' in dat.properties
    assert 'localization_density_obb' in dat.properties
    assert dat.region is None
    dat.region = Rectangle()
    assert caplog.record_tuples == [('locan.data.locdata', 30, 'Not all coordinates are within region.')]
    dat.region = dat.bounding_box.region
    assert repr(dat.region) == repr(dat.bounding_box.region)
    assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)
    assert 'region_measure_as' in dat.properties
    assert 'localization_density_as' in dat.properties

    assert len(dat.inertia_moments.eigenvalues) == 2
    assert 'orientation_im' in dat.properties
    assert 'circularity_im' in dat.properties


def test_LocData_empty(df_empty):
    dat = LocData()
    assert dat.data.empty
    assert dat.properties == {'localization_count': 0, 'region_measure_bb': 0}
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    assert dat.dimension == 0

    dat = LocData(dataframe=df_empty)
    assert dat.data.empty
    assert dat.properties == {'localization_count': 0, 'region_measure_bb': 0}
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    assert dat.dimension == 0
    # hulls are tested with locdata fixtures further down.


def test_LocData_from_dataframe(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    # print(dat.properties.keys())
    assert dat.bounding_box is not None
    assert list(dat.properties.keys()) == ['localization_count', 'position_x', 'position_y', 'region_measure_bb',
                                           'localization_density_bb', 'subregion_measure_bb']
    assert len(dat) == 5
    assert dat.meta.comment == COMMENT_METADATA.comment


def test_LocData_from_dataframe_empty(df_empty):
    dat = LocData.from_dataframe(dataframe=df_empty)
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    # print(dat.data)


def test_LocData_count(df_simple):
    # The following is commented out because it requires time and is not needed to ensure correct functionality.
    # import gc
    # print("Current number of LocData instances: ",
    #       len([item for item in gc.get_referrers(LocData) if isinstance(item, LocData)]))
    n_instances = LocData.count

    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    assert LocData.count == 1 + n_instances
    dat_2 = LocData.from_dataframe(dataframe=df_simple)
    assert(dat.properties == dat_2.properties)
    assert LocData.count == 2 + n_instances
    del dat
    assert LocData.count == 1 + n_instances


def test_LocData_from_dataframe_with_meta_dict(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta={'comment': 'some user comment'})
    assert (len(dat) == 5)
    assert (dat.references is None)
    assert (dat.meta.comment == COMMENT_METADATA.comment)


def test_LocData_from_coordinates():
    coordinates =[(200, 500), (200, 600), (900, 650), (1000, 600)]
    dat = LocData.from_coordinates(coordinates=coordinates, meta=COMMENT_METADATA)
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ['position_x', 'position_y'] for item in dat.coordinate_labels)
    assert len(dat) == 4
    assert dat.meta.comment == COMMENT_METADATA.comment

    coordinates =np.array([(200, 500), (200, 600), (900, 650), (1000, 600)])
    dat = LocData.from_coordinates(coordinates=coordinates, meta=COMMENT_METADATA)
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ['position_x', 'position_y'] for item in dat.coordinate_labels)
    assert len(dat) == 4

    dat = LocData.from_coordinates(coordinates=coordinates, coordinate_labels=['position_x', 'position_z'],
                                   meta=dict(comment='special order'))
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ['position_x', 'position_z'] for item in dat.coordinate_labels)
    assert len(dat) == 4

    with pytest.raises(ValueError):
        LocData.from_coordinates(coordinates=coordinates, coordinate_labels=['position_x', 'whatever'])

    dat = LocData.from_coordinates(coordinates=[])
    assert len(dat) == 0


def test_LocData_from_selection(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple)

    sel = LocData.from_selection(locdata=dat, indices=slice(1, 3))
    assert all(sel.data.index == [1, 2, 3])

    sel = LocData.from_selection(locdata=dat)
    assert len(sel) == len(df_simple)

    sel = LocData.from_selection(locdata=dat, indices=slice(4, 10))
    assert all(sel.data.index == [4])

    sel = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    assert (len(sel) == 3)
    assert (sel.references is dat)
    assert (sel.meta.comment == COMMENT_METADATA.comment)
    assert dat.bounding_box.region_measure != sel.bounding_box.region_measure
    with pytest.raises(AttributeError):
        sel.data = df_simple

    sel_sel = LocData.from_selection(locdata=sel, indices=[1, 3], meta={'comment': 'Selection of a selection.'})
    assert sel_sel.meta.comment == 'Selection of a selection.'
    assert (len(sel_sel) == 2)
    assert (sel_sel.references is sel)

    sel_empty = LocData.from_selection(locdata=dat, indices=[], meta={'comment': 'This selection is empty.'})
    assert sel_empty.data.empty
    assert len(sel_empty) == 0

    dat = LocData()
    sel_empty = LocData.from_selection(locdata=dat, indices=[])
    assert sel_empty.data.empty
    assert len(sel_empty) == 0

    sel_empty = LocData.from_selection(locdata=dat, indices=[1, 3])
    assert sel_empty.data.empty
    assert len(sel_empty) == 0


def test_LocData_from_collection(df_simple):
    col = LocData.from_collection([LocData(), LocData()], meta=COMMENT_METADATA)
    assert col.dimension == 0
    assert len(col) == 2
    assert col.properties == {'localization_count': 2, 'region_measure_bb': 0}
    assert (col.meta.comment == COMMENT_METADATA.comment)

    dat = LocData.from_dataframe(dataframe=df_simple)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.from_collection([sel_1, sel_2], meta=dict(comment='yet another comment'))
    assert (len(col.references) == 2)
    assert (len(col) == 2)
    assert (col.meta.comment == 'yet another comment')
    assert set(col.data.columns) == {'localization_count', 'localization_density_bb', 'position_x', 'position_y',
                                     'region_measure_bb', 'subregion_measure_bb'}

    with pytest.warns(UserWarning):
        col.update_convex_hulls_in_references()
    assert set(col.data.columns) == {'localization_count', 'localization_density_bb', 'position_x', 'position_y',
                                     'region_measure_bb', 'subregion_measure_bb',
                                     'region_measure_ch', 'localization_density_ch'}

    col.update_oriented_bounding_box_in_references()
    assert set(col.data.columns) == {'circularity_obb', 'localization_count', 'localization_density_bb',
                                     'localization_density_ch',
                                     'localization_density_obb', 'orientation_obb', 'position_x', 'position_y',
                                     'region_measure_bb', 'region_measure_ch', 'region_measure_obb',
                                     'subregion_measure_bb'}
    # print(col.properties)

    with pytest.warns(UserWarning):
        col.update_alpha_shape_in_references(alpha=1)
    assert set(col.data.columns) == {'circularity_obb', 'localization_count', 'localization_density_bb',
                                     'localization_density_ch',
                                     'localization_density_obb', 'orientation_obb', 'position_x', 'position_y',
                                     'region_measure_bb', 'region_measure_ch', 'region_measure_obb',
                                     'subregion_measure_bb',
                                     'region_measure_as', 'localization_density_as'}

    col.update_inertia_moments_in_references()
    assert set(col.data.columns) == {'circularity_obb', 'localization_count', 'localization_density_bb',
                                     'localization_density_ch',
                                     'localization_density_obb', 'orientation_obb', 'position_x', 'position_y',
                                     'region_measure_bb', 'region_measure_ch', 'region_measure_obb',
                                     'subregion_measure_bb',
                                     'region_measure_as', 'localization_density_as',
                                     'circularity_im', 'orientation_im'}


def test_LocData_selection_from_collection(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple)
    sel = []
    col = LocData.from_collection(sel)
    assert len(col) == 0

    for i in range(4):
        sel.append(LocData.from_selection(locdata=dat, indices=[i]))
    col = LocData.from_collection(sel, meta=COMMENT_METADATA)
    assert (len(col) == 4)
    assert (len(col.references) == 4)
    # print(col.data)

    col_sel = LocData.from_selection(locdata=col, indices=[0, 2, 3])
    assert (len(col_sel) == 3)
    assert(col_sel.references is col)
    # print(col_sel.data)

    col_sel_sel = LocData.from_selection(locdata=col_sel, indices=[2])
    assert (len(col_sel_sel) == 1)
    assert(col_sel_sel.references is col_sel)
    # print(col_sel_sel.data)


def test_LocData_concat(df_simple):
    col = LocData.concat([LocData(), LocData()], meta=COMMENT_METADATA)
    assert col.dimension == 0
    assert len(col) == 0
    assert col.properties == {'localization_count': 0, 'region_measure_bb': 0}
    assert (col.meta.comment == COMMENT_METADATA.comment)

    dat = LocData.from_dataframe(dataframe=df_simple)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.concat([sel_1, sel_2], meta=dict(comment='yet another comment'))
    assert (len(col) == 5)
    assert (col.meta.comment == 'yet another comment')
    assert len(col.references) == 2
    sel_1.reduce()
    col = LocData.concat([sel_1, sel_2], meta=COMMENT_METADATA)
    assert len(col.references) == 2
    sel_2.reduce()
    col = LocData.concat([sel_1, sel_2], meta=COMMENT_METADATA)
    assert col.references is None


def test_LocData_reduce(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel_1 = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    sel_2 = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    sel_1.reduce()
    assert sel_1.references is None
    assert (len(sel_1) == 3)
    assert (len(sel_1.data) == len(sel_2.data))
    col = LocData.from_collection([sel_1, sel_2], meta=COMMENT_METADATA)
    col.reduce()
    assert (len(col) == 2)
    assert col.references is None


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 1),
])
def test_locdata_reduce_empty(
        locdata_empty, locdata_single_localization,
        fixture_name, expected):
    dat = eval(fixture_name)
    new_dat = dat.reduce()
    assert len(new_dat) == expected


def test_LocData_reset(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    assert dat.properties['position_x'] == 2
    assert dat.properties['region_measure_bb'] == 20
    assert (dat.meta.comment == COMMENT_METADATA.comment)

    dat.dataframe['position_x'] = [0, 1, 4, 0, 0]
    dat.dataframe['position_y'] = [1, 3, 4, 0, 0]
    dat.reset()
    assert dat.properties['position_x'] == 1
    assert dat.properties['region_measure_bb'] == 16
    assert (dat.meta.comment == COMMENT_METADATA.comment)


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 1),
])
def test_locdata_reset_empty(
        locdata_empty, locdata_single_localization,
        fixture_name, expected):
    dat = eval(fixture_name)
    new_dat = dat.reset()
    assert len(new_dat) == expected


def test_LocData_update(df_simple, caplog):
    new_dataframe = pd.DataFrame.from_dict(
        {'position_x': [10, 0, 1, 4],
         'position_y': [10, 1, 3, 4],
         'frame': [0, 1, 1, 4]}
    )

    dat = LocData()
    dat.update(new_dataframe, meta=COMMENT_METADATA)
    pd.testing.assert_frame_equal(dat.data, new_dataframe)
    assert dat.properties['position_x'] == 3.75
    assert dat.meta.element_count == 4
    assert dat.meta.history[-1].name == "LocData.update"

    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    assert dat.properties['position_x'] == 2
    assert dat.meta.element_count == 5

    dat.update(new_dataframe, meta=dict(comment='yet another comment'))
    pd.testing.assert_frame_equal(dat.data, new_dataframe)
    assert dat.properties['position_x'] == 3.75
    assert dat.meta.element_count == 4
    assert dat.meta.history[-1].name == "LocData.update"

    sel = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    sel.update(new_dataframe, reset_index=True)
    assert caplog.record_tuples.pop() == ('locan.data.locdata', 30,
                                          'LocData.reduce() was applied since self.references was not None.')
    pd.testing.assert_frame_equal(sel.data, new_dataframe)
    assert sel.properties['position_x'] == 3.75
    assert sel.meta.element_count == 4
    assert sel.meta.history[-1].name == "LocData.update"


    sel = LocData.from_selection(locdata=dat, indices=[1, 2, 3, 4], meta=COMMENT_METADATA)
    sel.update(new_dataframe, reset_index=True)
    assert caplog.record_tuples.pop() == ('locan.data.locdata', 30,
                                          'LocData.reduce() was applied since self.references was not None.')
    pd.testing.assert_frame_equal(sel.data, new_dataframe)
    assert sel.properties['position_x'] == 3.75
    assert sel.meta.element_count == 4
    assert sel.meta.history[-1].name == "LocData.update"


def test_LocData_projection(locdata_3d):
    locdata = copy.deepcopy(locdata_3d)

    # test with single coordinate_label
    new_coordinate_labels = 'position_x'
    new_locdata = locdata.projection(coordinate_labels=new_coordinate_labels)

    assert all(label in new_coordinate_labels for label in new_locdata.coordinate_labels)
    assert new_locdata.dimension == 1
    assert len(new_locdata) == len(locdata)

    # test with multiple coordinate_labels
    new_coordinate_labels = ['position_x', 'position_y']
    new_locdata = locdata.projection(coordinate_labels=new_coordinate_labels)

    assert all(label in new_coordinate_labels for label in new_locdata.coordinate_labels)
    assert new_locdata.dimension == len(new_coordinate_labels)
    assert len(new_locdata) == len(locdata)


# locdata with added columns

def test_LocData_add_column_to_dataframe(df_simple):
    # from dataframe
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    dat.dataframe = dat.dataframe.assign(new=np.arange(5))
    assert(dat.data.equals(dat.dataframe))

    # from selection
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel.dataframe = sel.dataframe.assign(new=np.arange(3))
    assert(list(sel.dataframe.columns) == ['new'])
    assert(list(sel.references.dataframe.columns) == ['position_x', 'position_y'])
    assert all(list(sel.data.columns == ['position_x', 'position_y', 'new']))
    sel.reduce()
    assert all(list(sel.dataframe.columns == ['position_x', 'position_y', 'new']))

    # sel_2 = LocData.from_selection(locdata=sel, indices=[0, 2])
    # print(sel_2.data)
    # print(sel_2.references.data)
    # print(sel_2.references.references.data)
    # sel_2.reduce()
    # print(sel_2.data)

    # from collection
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.from_collection([sel_1, sel_2], meta=COMMENT_METADATA)
    col.dataframe = col.dataframe.assign(new=np.arange(2))
    # print(col.data.columns)
    assert all(list(col.data.columns == ['localization_count', 'position_x', 'position_y', 'region_measure_bb',
                                         'localization_density_bb', 'subregion_measure_bb', 'new']))


# locdata and metadata

def test_LocData_handling_metadata(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    # print(dat.meta)
    # print(dir(metadata_pb2))
    assert dat.meta.identifier
    assert dat.meta.creation_time
    assert dat.meta.source == metadata_pb2.DESIGN
    assert dat.meta.state == metadata_pb2.RAW
    assert dat.meta.history[0].name == "LocData.from_dataframe"
    assert dat.meta.element_count == 5

    dat.meta.comment = 'new comment'
    assert dat.meta.comment == 'new comment'

    dat.meta.map['variable key'] = 'new comment'
    assert dict(dat.meta.map) == {'variable key': 'new comment'}

    dat.meta.map['key_2'] = 'value_2'
    assert dat.meta.map['key_2'] == 'value_2'

    dat = LocData.from_dataframe(dataframe=df_simple, meta=dict(comment='yet another comment'))
    assert dat.meta.comment == 'yet another comment'


# locdata and regions

def test_locdata_region(df_simple):
    region = Rectangle((0, 0), 2, 1, 10)
    dat = LocData.from_dataframe(dataframe=df_simple)
    dat.region = region
    assert isinstance(dat.region, Region)
    assert dat.region.region_measure == 2


# standard LocData fixtures

@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 1),
    ('locdata_2d', 6),
    ('locdata_non_standard_index', 6)
])
def test_standard_locdata_objects(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_non_standard_index,
        fixture_name, expected):
    dat = eval(fixture_name)
    assert len(dat) == expected


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 0),
])
def test_locdata_hulls(
        locdata_empty, locdata_single_localization,
        fixture_name, expected):
    dat = eval(fixture_name)
    assert dat.bounding_box.region_measure == 0
    assert dat.oriented_bounding_box.region_measure == 0
    with pytest.warns(UserWarning):
        assert dat.convex_hull is None
    with pytest.raises(AttributeError):
        assert dat.region_measure_ch
    if len(dat) == 0:
        assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)
    else:
        with pytest.warns(UserWarning):
            assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 0),
])
def test_locdata_from_selection_exceptions(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_non_standard_index,
        fixture_name, expected):
    dat = eval(fixture_name)
    new_dat = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    assert len(new_dat) == expected


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_2d', 3),
    ('locdata_non_standard_index', 3)
])
def test_locdata_from_selection_(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_non_standard_index,
        fixture_name, expected):
    dat = eval(fixture_name)
    sel = LocData.from_selection(locdata=dat, indices=[1, 2, 5], meta=COMMENT_METADATA)
    assert (len(sel) == expected)
    assert set(sel.data.index) == set([1, 2, 5])
    # in coming versions: assert list(sel.data.index) == [2, 1, 5]
    assert (sel.references is dat)
    assert (sel.meta.comment == COMMENT_METADATA.comment)


def test_locdata_from_chunks_(locdata_non_standard_index):
    with pytest.raises(ValueError):
        chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index)

    chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index, chunks=((2, 1), (0, 3)))
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index, chunk_size=2)
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index, n_chunks=3)
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index, n_chunks=3, order='alternating')
    assert all(chunk_collection.references[0].data.index == [2, 10])


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', (0, (0,), 0, (0,))),
    ('locdata_single_localization', (1, (1,), 1, (1,))),
    ('locdata_2d', (3, (2, 2, 2), 1, (4,))),
    ('locdata_non_standard_index', (3, (2, 2, 2), 1, (4,)))
])
def test_locdata_from_chunks(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_non_standard_index,
        fixture_name, expected):
    dat = eval(fixture_name)
    chunk_collection = LocData.from_chunks(locdata=dat, chunk_size=2, meta=COMMENT_METADATA)
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[0]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[1])
    assert chunk_collection.meta.comment == COMMENT_METADATA.comment

    chunk_collection = LocData.from_chunks(locdata=dat, chunk_size=2, order='alternating',
                                           meta=dict(comment='yet another comment'))
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[0]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[1])
    assert chunk_collection.meta.comment == 'yet another comment'

    chunk_collection = LocData.from_chunks(locdata=dat, chunk_size=4, drop=True)
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[2]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[3])


def test_copy():
    locdata = LocData.from_dataframe(dataframe=pd.DataFrame({'col_1': [1, 2, 3], 'col_2': ['a', 'b', 'c']}))
    selection = LocData.from_selection(locdata, indices=[0, 1])
    new_locdata = copy.copy(selection)

    assert new_locdata is not selection
    assert len(selection) == len(new_locdata)
    for attr in ['dataframe', 'indices', 'references']:
        assert getattr(selection, attr) is getattr(new_locdata, attr)
    for attr in ['properties', 'data', 'meta']:
        assert getattr(selection, attr) is not getattr(new_locdata, attr)
    pd.testing.assert_frame_equal(new_locdata.dataframe, selection.dataframe)
    pd.testing.assert_frame_equal(new_locdata.data, selection.data)
    assert new_locdata.properties == selection.properties
    assert new_locdata.indices == selection.indices
    assert new_locdata.references == selection.references
    assert new_locdata.meta != selection.meta
    assert int(new_locdata.references.meta.identifier) == int(selection.meta.identifier) - 1
    assert int(new_locdata.meta.identifier) == int(selection.meta.identifier) + 1
    assert new_locdata.meta.creation_time == selection.meta.creation_time


def test_deepcopy():
    locdata = LocData.from_dataframe(dataframe=pd.DataFrame({'col_1': [1, 2, 3], 'col_2': ['a', 'b', 'c']}))
    selection = LocData.from_selection(locdata, indices=[0, 1])
    new_locdata = copy.deepcopy(selection)

    assert new_locdata is not selection
    assert len(selection) == len(new_locdata)
    for attr in ['data', 'dataframe', 'indices', 'meta', 'properties', 'references']:
        assert getattr(selection, attr) is not getattr(new_locdata, attr)
    pd.testing.assert_frame_equal(new_locdata.dataframe, selection.dataframe)
    pd.testing.assert_frame_equal(new_locdata.data, selection.data)
    assert new_locdata.properties == selection.properties
    assert new_locdata.indices == selection.indices
    assert new_locdata.references != selection.references
    assert new_locdata.meta != selection.meta
    assert int(new_locdata.references.meta.identifier) == int(selection.meta.identifier) + 1
    assert int(new_locdata.meta.identifier) == int(selection.meta.identifier) + 2
    assert new_locdata.meta.creation_time == selection.meta.creation_time


def test_locdata_print_summary(capfd, locdata_2d):
    locdata_2d.print_summary()
    captured = capfd.readouterr()
    for element in ['identifier', 'creation_time', 'source', 'state', 'element_count']:
        assert element in captured.out


def test_locdata_print_meta(capfd, locdata_2d):
    locdata_2d.print_meta()
    captured = capfd.readouterr()
    for element in ['identifier', 'creation_time', 'source', 'state', 'element_count']:
        assert element in captured.out
